import numpy as np
import nibabel as nib
import scipy.ndimage
from itertools import product

from .exceptions import NeuroLangException
from .aabb_tree import AABB, Tree, aabb_from_vertices, Node

__all__ = [
    'region_union', 'region_intersection', 'region_difference',
    'region_set_from_masked_data', 'take_principal_regions',
    'Region', 'VolumetricBrainRegion', 'PointSet',
    'ImplicitVBR', 'ExplicitVBR', 'ExplicitVBROverlay',
    'SphericalVolume', 'PlanarVolume'
]


def region_union(region_set, affine=None):
    region_set = (
        region for region in region_set
        if not isinstance(region, EmptyRegion)
    )
    return region_set_algebraic_op(set.union, region_set, affine)


def region_intersection(region_set, affine=None):
    if any(isinstance(region, EmptyRegion) for region in region_set):
        return EmptyRegion()    
    return region_set_algebraic_op(set.intersection, region_set, affine)


def region_difference(region_set, affine=None):
    return region_set_algebraic_op(set.difference, region_set, affine)


def region_set_algebraic_op(op, region_set, affine=None, n_dim=3):

    rs = list(filter(lambda x: x is not None, region_set))
    if affine is None:
        affine = next(iter(rs)).affine

    max_dim = (0,) * n_dim
    voxels_set_of_regions = []
    for region in rs:
        if not isinstance(region, ExplicitVBR):
            region = region.to_explicit_vbr(affine, max_dim)

        if (region.image_dim is not None and
                any(map(lambda x, y: x > y, region.image_dim, max_dim))):
            max_dim = region.image_dim
        voxels_set_of_regions.append(set(map(tuple, region.voxels)))

    result_voxels = np.array(tuple(op(*voxels_set_of_regions)))
    if len(result_voxels) == 0:
        return EmptyRegion()
    return ExplicitVBR(result_voxels, affine, max_dim)


class Region(object):
    def __init__(self, lb, ub):
        if not np.all([lb[i] < ub[i] for i in range(len(lb))]):
            raise NeuroLangException(
                'Lower bounds must be lower'
                ' than upper bounds when creating rectangular regions')
        self._bounding_box = AABB(lb, ub)

    def __hash__(self):
        return hash(self.bounding_box.limits.tobytes())

    @property
    def bounding_box(self):
        return self._bounding_box

    @property
    def center(self):
        return self.bounding_box.center

    @property
    def width(self):
        return self.bounding_box.width

    def __eq__(self, other):
        return self.bounding_box == other.bounding_box

    def __repr__(self):
        return f'Region(AABB={self.bounding_box})'

    @staticmethod
    def from_spatial_image_label(spatial_image, label, **kwargs):
        data = np.asanyarray(spatial_image.dataobj)
        voxels = np.transpose((data == label).nonzero())
        if 'image_dim' not in kwargs:
            kwargs['image_dim'] = spatial_image.shape

        if len(voxels) > 0:
            return ExplicitVBR(voxels, spatial_image.affine, **kwargs)
        else:
            return EmptyRegion()


class EmptyRegion(Region):
    def __init__(self):
        pass

    def __hash__(self):
        return hash(tuple())

    def __eq__(self, other):
        return isinstance(other, EmptyRegion)

    def __repr__(self):
        return 'EmptyRegion'


class VolumetricBrainRegion(Region):
    def to_xyz(self, affine):
        """return world coordinates of the region
         corresponding to the affine matrix transform"""
        raise NotImplementedError()

    def to_ijk(self, affine):
        """return ijk voxels coordinates
         corresponding to the affine matrix transform"""
        raise NotImplementedError()

    def to_xyz_set(self, affine):
        raise NotImplementedError()

    def to_ijk_set(self, affine):
        raise NotImplementedError()

    def remove_empty_bounding_boxes(self):
        raise NotImplementedError()

    def to_explicit_vbr(self, affine, image_shape):
        voxels_coordinates = self.to_ijk(affine)
        return ExplicitVBR(voxels_coordinates, affine, image_shape)


class PointSet(VolumetricBrainRegion):
    def __init__(
        self, points_ijk, affine_matrix, image_dim=None, prebuild_tree=False
    ):
        self.points_ijk = np.asanyarray(points_ijk, dtype=float)
        self.affine = affine_matrix
        self.affine_inv = np.linalg.inv(self.affine)
        self.image_dim = image_dim
        self._aabb_tree = None
        self._bounding_box = self.generate_bounding_box(self.points_ijk)
        if prebuild_tree:
            self.build_tree()

    @property
    def bounding_box(self):
        return self._bounding_box

    @property
    def aabb_tree(self):
        if self._aabb_tree is None:
            self.build_tree()
        return self._aabb_tree

    def generate_bounding_box(self, points_ijk):
        points_xyz = nib.affines.apply_affine(
            self.affine, points_ijk
        )
        return aabb_from_vertices(points_xyz)

    def build_tree(self):
        box = self._bounding_box
        tree = Tree()
        tree.add(box)

        affine_matrix_inv = np.linalg.inv(self.affine)

        stack = list([(box.lb, box.ub, self.points_ijk, tree.root)])
        while stack:
            node = stack.pop()
            lb, ub, parent_voxels, tree_node = node

            # this only works if the affine matrix is diagonal
            middle_point_xyz = (lb + ub) / 2
            middle_point_ijk = nib.affines.apply_affine(
                affine_matrix_inv, middle_point_xyz
            )

            axes = np.argsort(ub - lb)
            for ax in axes:
                left_mask = parent_voxels.T[ax] <= middle_point_ijk[ax]
                points_left = parent_voxels[left_mask]
                if 0 < len(points_left) < len(parent_voxels):
                    break
            else:
                continue

            height = tree_node.height - 1

            box_left = self.generate_bounding_box(points_left)
            tree_node.left = Node(
                box=box_left, parent=tree_node, height=height
            )
            stack.append(
                (box_left.lb, box_left.ub, points_left, tree_node.left)
            )

            points_right = parent_voxels[~left_mask]
            box_right = self.generate_bounding_box(points_right)
            tree_node.right = Node(
                box=box_right, parent=tree_node, height=height
            )
            stack.append(
                (box_right.lb, box_right.ub, points_right, tree_node.right)
            )

            tree.height = -height

        stack = [tree.root]
        while stack:
            node = stack.pop()
            node.height = tree.height + node.height
            if node.left is not None:
                stack.append(node.left)
            if node.right is not None:
                stack.append(node.right)

        self._aabb_tree = tree
        return tree

    def to_xyz(self, affine=None):
        if affine is None or np.allclose(affine, self.affine):
            affine = self.affine
        else:
            affine = np.dot(affine, np.linalg.inv(self.affine))
        return nib.affines.apply_affine(affine, self.points_ijk)

    def to_ijk(self, affine):
        return nib.affines.apply_affine(
            np.linalg.solve(affine, self.affine),
            self.points_ijk)

    def spatial_image(self, out=None, value=1):
        if out is None:
            mask = np.zeros(self.image_dim, dtype=np.int16)
            out = nib.spatialimages.SpatialImage(mask, self.affine)
        elif (out.shape != self.image_dim and
              not np.allclose(out.affine, self.affine)):
            raise ValueError("Image data has incompatible dimensionality")
        else:
            mask = np.asanyarray(out.dataobj)

        discrete_points_ijk = np.round(self.points_ijk).astype(int)
        mask[tuple(discrete_points_ijk.T)] = value
        return out

    def __eq__(self, other):
        return (np.array_equiv(self.affine, other.affine) and
                np.array_equiv(self.points_ijk, other.voxels))

    def __repr__(self):
        return (
            f'Region(PointSet= affine:{self.affine},'
            f' points_ijk:{self.points_ijk})'
        )

    def __hash__(self):
        return hash((self.points_ijk.tobytes(), self.affine.tobytes()))


class ExplicitVBR(VolumetricBrainRegion):
    def __init__(
        self, voxels, affine_matrix, image_dim=None, prebuild_tree=False
    ):
        self.voxels = np.asanyarray(voxels, dtype=int)
        self.affine = affine_matrix
        self.affine_inv = np.linalg.inv(self.affine)
        for ar in (self.voxels, self.affine, self.affine_inv):
            ar.setflags(write=False)
        self.image_dim = image_dim
        self._aabb_tree = None
        self._bounding_box = self.generate_bounding_box(self.voxels)
        if prebuild_tree:
            self.build_tree()

    @property
    def bounding_box(self):
        return self._bounding_box

    @property
    def aabb_tree(self):
        if self._aabb_tree is None:
            self.build_tree()
        return self._aabb_tree

    def generate_bounding_box(self, voxels_ijk):
        voxels_xyz = nib.affines.apply_affine(
            self.affine, np.concatenate((voxels_ijk, voxels_ijk + 1), axis=0)
        )
        return aabb_from_vertices(voxels_xyz)

    def build_tree(self):
        box = self._bounding_box
        tree = Tree()
        tree.add(box)

        affine_matrix_inv = np.linalg.inv(self.affine)

        stack = list([(box.lb, box.ub, self.voxels, tree.root)])
        while stack:
            node = stack.pop()
            lb, ub, parent_voxels, tree_node = node

            # this only works if the affine matrix is diagonal
            middle_point_xyz = (lb + ub) / 2
            middle_point_ijk = nib.affines.apply_affine(
                affine_matrix_inv, middle_point_xyz
            )

            axes = np.argsort(ub - lb)
            for ax in axes:
                left_mask = parent_voxels.T[ax] <= middle_point_ijk[ax]
                voxels_left = parent_voxels[left_mask]
                if 0 < len(voxels_left) < len(parent_voxels):
                    break
            else:
                continue

            height = tree_node.height - 1

            box_left = self.generate_bounding_box(voxels_left)
            tree_node.left = Node(
                box=box_left, parent=tree_node, height=height
            )
            stack.append(
                (box_left.lb, box_left.ub, voxels_left, tree_node.left)
            )

            voxels_right = parent_voxels[~left_mask]
            box_right = self.generate_bounding_box(voxels_right)
            tree_node.right = Node(
                box=box_right, parent=tree_node, height=height
            )
            stack.append(
                (box_right.lb, box_right.ub, voxels_right, tree_node.right)
            )

            tree.height = -height

        stack = [tree.root]
        while stack:
            node = stack.pop()
            node.height = tree.height + node.height
            if node.left is not None:
                stack.append(node.left)
            if node.right is not None:
                stack.append(node.right)

        self._aabb_tree = tree
        return tree

    def to_xyz(self, affine=None):
        if affine is None or np.allclose(affine, self.affine):
            affine = self.affine
        else:
            affine = np.dot(affine, np.linalg.inv(self.affine))
        return nib.affines.apply_affine(affine, self.voxels)

    def to_ijk(self, affine):
        return nib.affines.apply_affine(
            np.linalg.solve(affine, self.affine),
            self.voxels)

    def spatial_image(self, out=None, value=1):
        if out is None:
            mask = np.zeros(self.image_dim, dtype=np.int16)
            out = nib.spatialimages.SpatialImage(mask, self.affine)
        elif (out.shape != self.image_dim and
              not np.allclose(out.affine, self.affine)):
            raise ValueError("Image data has incompatible dimensionality")
        else:
            mask = np.asanyarray(out.dataobj)

        mask[tuple(self.voxels.T)] = value
        return out

    def __eq__(self, other):
        return (np.array_equiv(self.affine, other.affine) and
                np.array_equiv(self.voxels, other.voxels))

    def __repr__(self):
        return f'Region(VBR= affine:{self.affine}, voxels:{self.voxels})'

    def __hash__(self):
        return hash((self.voxels.tobytes(), self.affine.tobytes()))


class ExplicitVBROverlay(ExplicitVBR):
    def __init__(
        self, voxels, affine_matrix, overlay,
        image_dim=None, prebuild_tree=False
    ):
        super().__init__(
            voxels, affine_matrix,
            image_dim=image_dim, prebuild_tree=prebuild_tree
        )
        self.overlay = np.atleast_2d(overlay)
        if self.overlay.shape == (1, len(self.voxels)):
            self.overlay = self.overlay.T
        self.overlay.setflags(write=False)
        if self.overlay.shape[0] != self.voxels.shape[0]:
            raise ValueError(
                'The length of the overlay must be '
                'the same as that of the voxels'
            )

    def spatial_image(self, out=None, background_value=0):
        if self.overlay.shape[1:] == (1,):
            image_dim = self.image_dim
        else:
            image_dim = self.image_dim + self.overlay.shape[1:]
        out = self._obtain_empty_spatial_image(
            out, image_dim, background_value
        )
        image_data = out.dataobj

        image_data[tuple(self.voxels.T)] = self.overlay.squeeze()
        return out

    def _obtain_empty_spatial_image(self, out, image_dim, background_value):
        if out is None:
            mask = np.zeros(
                image_dim,
                dtype=self.overlay.dtype
            )
            out = nib.spatialimages.SpatialImage(mask, self.affine)
        elif (
            out.shape != image_dim and
            not np.allclose(out.affine, self.affine) and
            self.overlay.dtype != out.dataobj.dtype
        ):
            raise ValueError(
                "Image data has incompatible dimensionality or type"
            )
        else:
            mask = np.asanyarray(out.dataobj)
            mask[:] = background_value
        return out

    def __eq__(self, other):
        return (
            isinstance(other, ExplicitVBROverlay) and
            np.array_equiv(self.affine, other.affine) and
            np.array_equiv(self.voxels, other.voxels) and
            np.array_equiv(self.overlay, other.overlay)
        )

    def __repr__(self):
        return (
            f'Region(VBR= affine:{self.affine}, '
            f'voxels:{self.voxels}, overlay:{self.overlay})'
        )

    def __hash__(self):
        return hash(
            (
                self.voxels.tobytes(),
                self.affine.tobytes(),
                self.overlay.tobytes()
            )
        )


def region_set_from_masked_data(data, affine, dim):
    s = scipy.ndimage.generate_binary_structure(3, 2)
    labeled_array, num_features = scipy.ndimage.label(data, structure=s)
    regions = set()
    for i in range(1, num_features):
        region_voxels = list(zip(*np.where(labeled_array == i)))
        regions.add(ExplicitVBR(region_voxels, affine, dim))

    return set(regions)


def take_principal_regions(region_set, k):
    sorted_by_size = sorted(list(region_set),
                            key=lambda x: len(x.voxels), reverse=True)
    return set(sorted_by_size[:k])


class ImplicitVBR(VolumetricBrainRegion):
    def __contains__(self, voxel):
        raise NotImplementedError()

    def to_ijk(self, affine):
        raise NotImplementedError()

    def to_xyz(self, affine):
        raise NotImplementedError()


class BoundigBoxSequenceElement(ImplicitVBR):
    def __init__(self, bounding_box_sequence):
        self.bounding_box_sequence = bounding_box_sequence
        self._bounding_box = self.generate_bounding_box(
            self.bounding_box_sequence
        )
        self._aabb_tree = None

    @staticmethod
    def generate_bounding_box(bounding_box_sequence):
        res = bounding_box_sequence[0]
        for bb in bounding_box_sequence[1:]:
            res = res.union(bb)
        return res

    @property
    def aabb_tree(self):
        if self._aabb_tree is None:
            self.build_tree()
        return self._aabb_tree

    def to_ijk(self, affine):
        midpoints_xyz = np.array([
            (b.ub + b.lb) / 2
            for b in self.bounding_box_sequence
        ])
        voxel_coordinates = np.round(nib.affines.apply_affine(
            np.linalg.inv(affine), midpoints_xyz
        ))
        return voxel_coordinates

    def build_tree(self):
        box = self._bounding_box
        tree = Tree()
        tree.add(box)

        stack = list([(box.lb, box.ub, self.bounding_box_sequence, tree.root)])
        while stack:
            node = stack.pop()
            lb, ub, parent_boxes, tree_node = node

            # this only works if the affine matrix is diagonal
            middle_point_xyz = (lb + ub) / 2

            axes = np.argsort(ub - lb)
            for ax in axes:
                ub_box_lower = ub.copy()
                ub_box_lower[ax] = middle_point_xyz[ax]
                box_lower = AABB(lb, ub_box_lower)

                boxes_lower = []
                boxes_higher = []
                for box in parent_boxes:
                    if box_lower.contains(box):
                        boxes_lower.append(box)
                    else:
                        boxes_higher.append(box)

                if 0 < len(boxes_lower) < len(parent_boxes):
                    break
            else:
                continue

            height = tree_node.height - 1
            box_left = self.generate_bounding_box(boxes_lower)
            tree_node.left = Node(
                box=box_left, parent=tree_node, height=height
            )
            stack.append(
                (box_left.lb, box_left.ub, boxes_lower, tree_node.left)
            )

            box_right = self.generate_bounding_box(boxes_higher)
            tree_node.right = Node(
                box=box_right, parent=tree_node, height=height
            )
            stack.append(
                (box_right.lb, box_right.ub, boxes_higher, tree_node.right)
            )
            tree.height = -height

        stack = [tree.root]
        while stack:
            node = stack.pop()
            node.height = tree.height + node.height
            if node.left is not None:
                stack.append(node.left)
            if node.right is not None:
                stack.append(node.right)

        return tree


class SphericalVolume(ImplicitVBR):
    def __init__(self, center, radius):
        self._center = np.asanyarray(center, dtype=int)
        self._radius = np.asanyarray(radius, dtype=int)
        lb = self._center - self._radius
        ub = self._center + self._radius
        self._bounding_box = AABB(lb, ub)

    @property
    def center(self):
        return self._center

    @property
    def radius(self):
        return self._radius

    @property
    def bounding_box(self):
        return self._bounding_box

    def to_ijk(self, affine):
        bb = self.bounding_box
        bounds_voxels = nib.affines.apply_affine(
            np.linalg.inv(affine), np.array([bb.lb, bb.ub]))
        ranges = [range(int(min(bounds_voxels[:, i])),
                        int(max(bounds_voxels[:, i])))
                  for i in range(bb.dim)]

        if all(r.start == r.stop for r in ranges):
            voxel_coordinates = np.array([[r.start for r in ranges]])
        else:
            voxel_coordinates = np.array([
                point for point in
                np.array(list(product(*ranges))) if
                nib.affines.apply_affine(affine, point) in self
            ])

        return voxel_coordinates

    def __contains__(self, point):
        point = np.atleast_2d(point)
        return np.all(
            np.linalg.norm(self._center - point, axis=1) <= self._radius
        )

    def __hash__(self):
        return hash(self.bounding_box.limits.tobytes())

    def __eq__(self, other):
        return (np.array_equiv(self._center, other._center) and
                self._radius == other._radius)

    def __repr__(self):
        return (f'SphericalVolume(Center={tuple(self._center)}, '
                f'Radius={self._radius})')


class PlanarVolume(ImplicitVBR):
    def __init__(self, origin, vector, direction=1, limit=1000):
        self._origin = np.array(origin)

        if not np.any([vector[i] > 0 for i in range(len(vector))]):
            raise ValueError('Vector normal to the plane must be non-zero')
        self._vector = np.array(vector) / np.linalg.norm(vector)

        if direction not in [1, -1]:
            raise ValueError('Direction must either be'
                             ' 1 (superior to) or -1 (inferior to)')
        self._dir = direction

        if limit <= 0:
            raise ValueError('Limit must be a positive value')
        self._limit = limit

        box_limit = np.asanyarray((self._dir * self._limit,) * 3, dtype=int)
        box_limit_in_plane = np.asanyarray(
            self.project_point_to_plane(box_limit), dtype=int) * -1
        lb, ub = sorted([box_limit, box_limit_in_plane], key=lambda x: x[0])
        self._bounding_box = AABB(lb, ub)

    def project_point_to_plane(self, point):
        point = np.array(point)
        dist = np.dot(point - self._origin, self._vector)
        return tuple(point - dist * self._vector)

    @property
    def bounding_box(self):
        return self._bounding_box

    def to_ijk(self, affine):
        bb = self.bounding_box
        bounds_voxels = nib.affines.apply_affine(np.linalg.inv(affine),
                                                 np.array([bb.lb, bb.ub]))
        ranges = [range(int(min(bounds_voxels[:, i])),
                        int(max(bounds_voxels[:, i])))
                  for i in range(bb.dim)]

        return np.array(list(product(*ranges)))

    def __contains__(self, point):
        point = np.atleast_2d(point)
        return np.all(
            np.sum(self._vector * (self._origin - point), axis=1) == 0
        )

    def __hash__(self):
        return hash(self.bounding_box.limits.tobytes())

    def __eq__(self, other):
        return (np.array_equiv(self._origin, other._origin) and
                np.array_equiv(self._vector, other._vector))

    def __repr__(self):
        return (f'PlanarVolume(Origin={tuple(self._origin)},'
                f' Normal Vector={self._vector})')



from rdflib import OWL, RDF, RDFS

from ... import expression_walker as ew
from ...expression_walker import ReplaceExpressionWalker
from ...expressions import Constant, ExpressionBlock, Symbol
from ...logic import (
    Conjunction,
    ExistentialPredicate,
    FunctionApplication,
    Implication,
)
from ...logic.transformations import CollapseConjunctions
from ..aggregation import DatalogWithAggregationMixin
from ..expressions import TranslateToLogic
from ..ontologies_parser import RightImplication
from ..ontologies_rewriter import OntologyRewriter

S_ = Symbol
C_ = Constant
EP_ = ExistentialPredicate
EB_ = ExpressionBlock
FA_ = FunctionApplication
I_ = Implication
RI_ = RightImplication


class DatalogTranslator(
    TranslateToLogic, ew.IdentityWalker, DatalogWithAggregationMixin
):
    pass


def test_normal_rewriting_step():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(a, db, b))

    qB = EB_((q,))
    sigmaB = EB_((sigma,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 2
    imp1 = rewrite.pop()
    imp2 = rewrite.pop()
    assert imp1[0] == q or imp2[0] == q
    q2 = I_(p(b), project(b) & inArea(b, db))
    q2 = dt.walk(q2)
    assert imp1[0] == q2 or imp2[0] == q2


def test_more_than_one_free_variable():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    w = S_("w")
    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    c = S_("c")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(w, z, y, x))
    q = I_(p(b), hasCollaborator(c, a, db, b))

    qB = EB_((q,))
    sigmaB = EB_((sigma,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 2
    imp1 = rewrite.pop()
    imp2 = rewrite.pop()
    assert imp1[0] == q or imp2[0] == q
    q2 = I_(p(b), project(b) & inArea(b, db))
    q2 = dt.walk(q2)
    assert imp1[0] == q2 or imp2[0] == q2


def test_unsound_rewriting_step_constant():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    b = S_("b")
    db = C_("db")
    c = C_("c")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(c, db, b))

    qB = EB_((q,))
    sigmaB = EB_((sigma,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    imp = rewrite.pop()
    assert imp[0] == q


def test_unsound_rewriting_step_shared():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    b = S_("b")
    db = C_("db")

    sigma = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    q = I_(p(b), hasCollaborator(b, db, b))

    qB = EB_((q,))
    sigmaB = EB_((sigma,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    imp = rewrite.pop()
    assert imp[0] == q


def test_outside_variable():
    s = S_("s")
    r = S_("r")
    t = S_("t")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")

    a = S_("a")
    b = S_("b")
    c = S_("c")
    e = S_("e")

    sigma = RI_(s(x) & r(x, y), t(x, y, z))
    q2 = I_(p(a), s(c) & t(a, b, c) & t(a, e, c))

    qB = EB_((q2,))
    sigmaB = EB_((sigma,))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 1
    factorized = [x for x in rewrite if x[1] == "f"]
    assert len(factorized) == 0


def test_example_4_3():
    project = S_("project")
    inArea = S_("inArea")
    hasCollaborator = S_("hasCollaborator")
    collaborator = S_("collaborator")
    p = S_("p")

    x = S_("x")
    y = S_("y")
    z = S_("z")
    a = S_("a")
    b = S_("b")
    c = S_("c")

    sigma1 = RI_(project(x) & inArea(x, y), hasCollaborator(z, y, x))
    sigma2 = RI_(hasCollaborator(x, y, z), collaborator(x))

    q = I_(p(b, c), hasCollaborator(a, b, c) & collaborator(a))

    qB = EB_((q,))
    sigmaB = EB_((sigma1, sigma2))

    dt = DatalogTranslator()
    qB = dt.walk(qB)
    sigmaB = dt.walk(sigmaB)

    orw = OntologyRewriter(qB, sigmaB)
    rewrite = orw.Xrewrite()

    assert len(rewrite) == 4


def test_infinite_walker():

    subClassOf = Symbol(str(RDFS.subClassOf))
    rest = Symbol(str(OWL.rest))
    reg = Symbol("reg")

    x1 = Symbol("x1")
    y1 = Symbol("y1")
    x = Symbol("x")
    y = Symbol("y")

    sigma_ant = Conjunction((subClassOf(x1, y1), rest(x1, y1), rest(y1, y1)))
    S = subClassOf(x1, y1)
    q_ant = Conjunction((subClassOf(x1, y1), rest(y, x1), rest(x, y1), reg(x)))

    replace = dict({S: sigma_ant})
    rsw = ReplaceExpressionWalker(replace)
    sigma_rep = rsw.walk(q_ant)
    sigma_rep = CollapseConjunctions().walk(sigma_rep)

    expected = Conjunction((sigma_ant, rest(y, x1), rest(x, y1), reg(x)))
    expected = CollapseConjunctions().walk(expected)

    assert sigma_rep == expected



from ..expressions import Definition, FunctionApplication, Symbol
from ..logic import BinaryLogicOperator

PROB = Symbol("PROB")


class ProbabilisticBinaryLogicOperator(BinaryLogicOperator):
    pass


class Condition(ProbabilisticBinaryLogicOperator):
    def __init__(self, conditioned, conditioning):
        self.conditioned = conditioned
        self.conditioning = conditioning
        self._symbols = conditioned._symbols | conditioning._symbols

    def __repr__(self):
        return f"[{self.conditioned} | {self.conditioning}]"


class ProbabilisticPredicate(Definition):
    def __init__(self, probability, body):
        self.probability = probability
        self.body = body
        self._symbols = body._symbols | self.probability._symbols

    def __repr__(self):
        return "ProbabilisticPredicate{{{} :: {} : {}}}".format(
            self.probability, self.body, self.type
        )


class Grounding(Definition):
    def __init__(self, expression, relation):
        self.expression = expression
        self.relation = relation

    def __repr__(self):
        return "Grounding{{{}}}\n{}".format(
            repr(self.expression), repr(self.relation)
        )


class ProbabilisticChoiceGrounding(Grounding):
    """
    Class used to differentiate the grounding of a probabilistic
    choice from the grounding of other choices.

    """


class ProbabilisticQuery(FunctionApplication):
    pass



from ....expressions import (
    Symbol,
    Constant,
    ExpressionBlock,
)
from ....expression_walker import ExpressionBasicEvaluator
from ....logic import Implication, Conjunction, UniversalPredicate
from ....datalog.basic_representation import DatalogProgram
from ....datalog.expressions import TranslateToLogic
from ....datalog.chase import (
    ChaseGeneral,
    ChaseMGUMixin,
    ChaseNaive,
)
from ..translate_to_dl import (
    TranslateToDatalog,
    TransformIntoConjunctionOfDatalogSentences,
)

from itertools import product


intersects = Symbol("intersects")
region = Symbol("region")
book = Symbol("book")
likes = Symbol("likes")
references = Symbol("references")
x = Symbol("X")
y = Symbol("Y")
Jones = Constant("Jones")
Ulysses = Constant("Ulysses")
Odyssey = Constant("Odyssey")


def test_translate_to_dl():
    ttdl = TranslateToDatalog()
    program = ttdl.translate_sentence("if Y intersects X then X intersects Y")

    assert program == ExpressionBlock(
        (Implication(intersects(x, y), intersects(y, x)),)
    )


def test_translate_to_dl_2():
    ttdl = TranslateToDatalog()
    program = ttdl.translate_sentence(
        "if a region Y intersects a region X then X intersects Y"
    )

    expected = ExpressionBlock(
        (
            Implication(
                intersects(x, y),
                Conjunction((intersects(y, x), region(x), region(y),)),
            ),
        )
    )
    assert program == expected


def test_distribute_implication_conjunctive_head():
    x = Symbol("x")
    y = Symbol("y")
    A = Symbol("A")
    B = Symbol("B")
    C = Symbol("C")

    exp = TransformIntoConjunctionOfDatalogSentences().walk(
        UniversalPredicate(
            x,
            UniversalPredicate(
                y, Implication(Conjunction((B(x), C(y),)), A(x, y),)
            ),
        )
    )

    assert exp == Conjunction(
        (
            UniversalPredicate(
                x, UniversalPredicate(y, Implication(B(x), A(x, y))),
            ),
            UniversalPredicate(
                x, UniversalPredicate(y, Implication(C(y), A(x, y))),
            ),
        )
    )


class Datalog(TranslateToLogic, DatalogProgram, ExpressionBasicEvaluator):
    pass


class Chase(ChaseNaive, ChaseMGUMixin, ChaseGeneral):
    pass


def test_solver_integration():
    """
    Given an extensional base having the predicates:
      `likes(x, y), book(x), references(x, y)`

    And the facts:
      `book(Odyssey), book(Ulysses)`

    Adding the next sentences to the base:
      `Jones likes a book if it references Odyssey.`
      `Ulysses references Odyssey.`

    Should answer:
      `Does Jones like Ulysses`
    """

    ttdl = TranslateToDatalog()
    program = ttdl.translate_block(
        """
        if a book X references Odyssey then Jones likes X.
        Ulysses references Odyssey.
        """
    )

    dl = Datalog()
    dl.add_extensional_predicate_from_tuples(book, {(Odyssey,), (Ulysses,)})
    dl.walk(program)

    dc = Chase(dl)
    solution = dc.build_chase_solution()

    assert (Jones, Ulysses) in solution[likes].value


def test_conjunctions():
    reaches = Symbol("reaches")
    edge = Symbol("edge")

    ttdl = TranslateToDatalog()
    program = ttdl.translate_block(
        """
        if `edge(X, Y)` then X reaches Y.
        if X reaches Y then Y reaches X.
        if A reaches B and B reaches C then A reaches C.
        """
    )
    dl = Datalog()
    dl.add_extensional_predicate_from_tuples(edge, {(1, 2), (2, 3), (3, 4)})
    dl.walk(program)

    dc = Chase(dl)
    solution = dc.build_chase_solution()

    for a, b in product((1, 2, 3, 4), (1, 2, 3, 4)):
        if a == b:
            continue
        assert (Constant(a), Constant(b)) in solution[reaches].value


def test_conjunctions_2():
    edge = Symbol("edge")
    start = Symbol("start")
    has = Symbol("has")
    program = TranslateToDatalog().translate_block(
        """
        if `edge(X, Y)` then X reaches Y.
        if X reaches Y then Y reaches X.
        if X reaches Y and X has "label" then Y has "label".
        if `start(X)` then X has "label".
        """
    )
    dl = Datalog()
    dl.add_extensional_predicate_from_tuples(
        edge, {("A", "B"), ("B", "C"), ("C", "D"), ("F", "G")}
    )
    dl.add_extensional_predicate_from_tuples(start, {("A",)})
    dl.walk(program)
    dc = Chase(dl)
    solution = dc.build_chase_solution()

    for x in ("A", "B", "C", "D"):
        assert (Constant(x), Constant("label")) in solution[has].value


def test_conjunction_3():
    edge = Symbol("edge")
    has = Symbol("has")
    program = TranslateToDatalog().translate_block(
        """
        if `edge(X, Y)` then X reaches Y.
        if X reaches Y and Y reaches Z then X reaches Z.
        if X reaches X then X has "cycle".
        """
    )
    dl = Datalog()
    dl.add_extensional_predicate_from_tuples(
        edge,
        {
            ("A", "B"),
            ("B", "C"),
            ("C", "D"),
            ("D", "E"),
            ("E", "B"),
            ("E", "F"),
        },
    )
    dl.walk(program)
    dc = Chase(dl)
    solution = dc.build_chase_solution()

    for x in ("B", "C", "D", "E"):
        assert (Constant(x), Constant("cycle")) in solution[has].value
    for x in ("A", "F"):
        assert (Constant(x), Constant("cycle")) not in solution[has].value



from ...logic.unification import most_general_unifier, apply_substitution
from ...logic.expression_processing import extract_logic_free_variables
from ...expressions import Symbol, Constant, Expression
from ...expression_walker import ReplaceSymbolWalker
from .exceptions import (
    ParseException,
    AmbiguousSentenceException,
    CouldNotParseException,
    TokenizeException,
    GrammarException,
)
from collections import namedtuple
import re


Quote = Symbol("Quote")
CODE_QUOTE = "`"
STRING_QUOTE = '"'


class Rule(Expression):
    def __init__(self, head, constituents):
        if not isinstance(constituents, tuple):
            raise GrammarException(
                "constituents must be a tuple of expressions"
            )
        self.head = head
        self.constituents = constituents
        self.is_root = False

    def __repr__(self):
        return (
            repr(self.head)
            + " --> "
            + ", ".join(repr(c) for c in self.constituents)
        )


class RootRule(Rule):
    def __init__(self, head, constituents):
        super().__init__(head, constituents)
        self.is_root = True


class Grammar(Expression):
    def __init__(self, rules):
        if not isinstance(rules, tuple) or any(
            not isinstance(r, Rule) for r in rules
        ):
            raise GrammarException("rules must be a tuple of Rule instances")
        self.rules = rules

    def __repr__(self):
        return (
            "Grammar {\n"
            + "\n".join("  " + repr(c) for c in self.rules)
            + "\n}"
        )


class Lexicon:
    def __init__(self):
        pass

    def get_meanings(self, word):
        raise NotImplementedError("")


class DictLexicon(Lexicon):
    """Convenience implementation of lexicon over dict"""

    def __init__(self, d):
        self.dict = d

    def get_meanings(self, token):
        if isinstance(token, Constant):
            if token.value in self.dict:
                return self.dict[token.value]
        return ()


class Tokenizer:
    def __init__(self, grammar, quotes=[CODE_QUOTE, STRING_QUOTE]):
        self.grammar = grammar
        self.matches = []
        for q in quotes:
            self.matches.append(
                (re.compile(f"^{q}.+?{q}"), self.yield_quote(q))
            )

        self.matches.append((re.compile("^[\\w\\-]+?\\b"), self.yield_word))
        self.matches.append((re.compile("^,\\s"), self.yield_comma))

    def yield_quote(self, q):
        def foo(span):
            return Quote(Constant(q), Constant[str](span[1:-1]))

        return foo

    def yield_word(self, span):
        return Constant[str](span)

    def yield_comma(self, span):
        return Constant[str](",")

    def tokenize(self, string):
        rem = string.strip()
        tokens = []
        while rem:
            t, rem = self.next_token(rem)
            tokens.append(t)
        return tokens

    def next_token(self, text):
        for r, on_match in self.matches:
            m = r.match(text)
            if m:
                e = m.end()
                span = text[:e].strip()
                rem = text[e:].lstrip()
                return on_match(span), rem

        raise TokenizeException(f"Couldnt match token at: {text}")


class Chart(list):
    pass


class ChartParser:
    Edge = namedtuple(
        "Edge", "head rule completed remaining used_edges unification"
    )

    def __init__(self, grammar, lexicon):
        self.grammar = grammar
        self.lexicon = lexicon
        self.tokenizer = Tokenizer(grammar)

    def recognize(self, string):
        try:
            self.parse(string)
        except ParseException:
            return False
        return True

    def parse(self, string):
        tokens = self.tokenizer.tokenize(string)
        self._fill_chart(tokens)
        compl = [
            e
            for e in self.chart[0][len(tokens)]
            if e.rule.is_root and not e.remaining
        ]
        results = [self._build_tree(e, e.unification) for e in compl]
        if len(results) == 0:
            raise CouldNotParseException(string)
        if len(results) > 1:
            raise AmbiguousSentenceException(string, results)

        return results

    def _build_tree(self, edge, unif):
        head = _lu.substitute(edge.head, unif)
        args = ()
        for ce in edge.used_edges:
            cu = ce.unification.copy()
            cu.update(unif)
            args += (self._build_tree(ce, cu),)

        if args:
            return head(*args)
        return head

    def _fill_chart(self, tokens):
        self._initialize(tokens)
        while self.agenda:
            self.agenda.sort(key=lambda e: (-e[1], -e[2]))
            edge, i, j = self.agenda.pop()
            self._predict(edge, i, j)
            self._complete(edge, i, j)

    # For every word wi add the edge
    #   [wi →  • , (i, i+1)]
    #
    def _initialize(self, tokens):
        self.agenda = []
        self.chart = Chart(
            [
                [[] for _ in range(len(tokens) + 1)]
                for _ in range(len(tokens) + 1)
            ]
        )
        for i, t in enumerate(tokens):
            word_edge = self.Edge(t, None, [], [], [], dict())
            self.chart[i][i + 1].append(word_edge)
            self.agenda.append((word_edge, i, i + 1))
            for m in self.lexicon.get_meanings(t):
                edge = self.Edge(
                    m, None, [word_edge.head], [], [word_edge], dict()
                )
                self.chart[i][i + 1].append(edge)
                self.agenda.append((edge, i, i + 1))

    # If the chart contains the complete edge
    #   [A → α • , (i, j)]
    # and the grammar contains the production
    #   B → A β
    # then add the self-loop edge
    #   [B →  • A β , (i, i)]
    #
    def _predict(self, edge, i, j):
        for rule in self.grammar.rules:
            if _lu.unify(rule.constituents[0], edge.head) and not any(
                rule == e.rule for e in self.chart[i][i]
            ):
                self.chart[i][i].append(self._create_edge_for_rule(rule))

    def _create_edge_for_rule(self, rule):
        fv = extract_logic_free_variables(rule.head)
        for c in rule.constituents:
            fv |= extract_logic_free_variables(c)
        rsw = ReplaceSymbolWalker({v: Symbol.fresh() for v in fv})
        nr = rsw.walk(rule)
        return self.Edge(nr.head, rule, [], nr.constituents, [], dict())

    # If the chart contains the edges
    #   [A → α • B β , (i, j)]
    #   [B → γ • , (j, k)]
    # then add the new edge
    #   [A → α B • β , (i, k)]
    # where α, β, and γ are (possibly empty) sequences
    # of terminals or non-terminals
    #
    def _complete(self, completed_edge, j, k):
        for i, e in self._uncompleted_edges_ending_at(j):
            u = _lu.unify(e.remaining[0], completed_edge.head)
            if not u:
                continue
            new_edge = self._complete_edge(e, completed_edge, u)
            self._add_completed_edge_to_chart(new_edge, i, k)

    def _uncompleted_edges_ending_at(self, j):
        for i in range(j + 1):
            for e in self.chart[i][j]:
                if e.remaining:
                    yield i, e

    def _complete_edge(self, edge_a, edge_b, u):
        # here completed could have the references
        #   to the involved edges
        n_completed = edge_a.completed + [u[1]]
        n_used_edges = edge_a.used_edges + [edge_b]
        n_remaining = edge_a.remaining[1:]
        unif = edge_a.unification.copy()
        unif.update(u[0])
        if n_remaining:
            n_remaining = [_lu.substitute(r, unif) for r in n_remaining]
            return self.Edge(
                edge_a.head,
                edge_a.rule,
                n_completed,
                n_remaining,
                n_used_edges,
                unif,
            )
        else:
            new_head = _lu.substitute(edge_a.head, unif)
            return self.Edge(
                new_head, edge_a.rule, n_completed, [], n_used_edges, unif
            )

    def _add_completed_edge_to_chart(self, new_edge, i, k):
        if not new_edge:
            return
        if not new_edge.remaining:
            self.agenda.append((new_edge, i, k))
        self.chart[i][k].append(new_edge)


class _lu:
    """
    Quick solution to get unification working without function applications.
    """

    @staticmethod
    def unify(a, b):
        w = Symbol("wrapper")
        u = most_general_unifier(w(a), w(b))
        if u is None:
            return
        return u[0], u[1].args[0]

    @staticmethod
    def substitute(exp, sub):
        w = Symbol("wrapper")
        n = apply_substitution(w(exp), sub)
        return n.args[0]



from collections.abc import Iterable
from fnmatch import fnmatch
from xml.etree import ElementTree


def parse_region_label_map(
    labeled_im, selected_labels=None
):
    extension_header = ElementTree.fromstring(
        labeled_im.header.extensions[0].get_content()
    )
    labels = extension_header.findall(".//Label")

    if selected_labels is None:
        labeltable = {
            l.text: int(l.get('Key'))
            for l in labels
            }
    elif isinstance(selected_labels, str):
        labeltable = {
            l.text: int(l.get('Key'))
            for l in labels
            if fnmatch(l.text, selected_labels)
        }
    elif isinstance(selected_labels, Iterable):
        labeltable = {
            l.text: int(l.get('Key'))
            for l in labels
            if l.text in selected_labels
            }

    if '???' in labeltable.keys():
        del labeltable['???']
    return labeltable



from ...expression_pattern_matching import add_match
from ...relational_algebra import Projection
from ...relational_algebra_provenance import (
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceCountingSolver,
)


class NoisyORProbabilityProvenanceSolver(
    RelationalAlgebraProvenanceCountingSolver
):
    @add_match(Projection)
    def projection(self, projection_op):
        prov_set = self.walk(projection_op.relation)
        prov_col = prov_set.provenance_column
        group_cols = list(col.value for col in projection_op.attributes)
        aggregations = {prov_col: lambda s: 1 - (1 - s).prod()}
        new_relation = prov_set.value
        new_relation = new_relation.aggregate(group_cols, aggregations)
        proj_cols = [prov_col] + group_cols
        new_relation = new_relation.projection(*proj_cols)
        return ProvenanceAlgebraSet(new_relation, prov_col)



from typing import AbstractSet

from ...expressions import Constant, Symbol
from ...relational_algebra import (
    ColumnStr,
    ConcatenateConstantColumn,
    NamedRelationalAlgebraFrozenSet,
    Projection,
    RelationalAlgebraSolver,
    RenameColumn,
    Selection,
    eq_,
    str2columnstr_constant,
)
from ...relational_algebra_provenance import ProvenanceAlgebraSet


def fresh_name_relation(ra_set):
    columns = tuple(Symbol.fresh().name for _ in range(ra_set.value.arity))
    relation = NamedRelationalAlgebraFrozenSet(columns, ra_set.value)
    return Constant[AbstractSet](relation)


def rename_columns_for_args_to_match(relation, src_args, dst_args):
    """
    Rename the columns of a relation so that they match the targeted args.

    Parameters
    ----------
    relation : ProvenanceAlgebraSet or RelationalAlgebraOperation
        The relation on which the renaming of the columns should happen.
    src_args : tuple of Symbols
        The predicate's arguments currently matching the columns.
    dst_args : tuple of Symbols
        New args that the naming of the columns should match.

    Returns
    -------
    RelationalAlgebraOperation
        The unsolved nested operations that apply the renaming scheme.

    """
    src_cols = list(str2columnstr_constant(arg.name) for arg in src_args)
    dst_cols = list(str2columnstr_constant(arg.name) for arg in dst_args)
    result = relation
    for dst_col in set(dst_cols):
        idxs = [i for i, c in enumerate(dst_cols) if c == dst_col]
        result = RenameColumn(result, src_cols[idxs[0]], dst_col)
        for idx in idxs[1:]:
            result = Selection(result, eq_(src_cols[idx], dst_col))
    return result


def build_always_true_provenance_relation(relation, prob_col=None):
    """
    Construct a provenance set from a relation with probabilities of 1
    for all tuples in the relation.

    The provenance column is named after the ``prob_col`` argument. If
    ``prob_col`` is already in the columns of the relation, it is
    removed before being re-added.

    Parameters
    ----------
    relation : NamedRelationalAlgebraFrozenSet
        The relation containing the tuples that will be in the
        resulting provenance set.
    prob_col : Constant[ColumnStr]
        Name of the provenance column that will contain constant
        probabilities of 1.

    Returns
    -------
    ProvenanceAlgebraSet

    """
    if prob_col is None:
        prob_col = ColumnStr(Symbol.fresh().name)
    # remove the probability column if it is already there
    elif prob_col in relation.value.columns:
        kept_cols = tuple(
            str2columnstr_constant(col)
            for col in relation.value.columns
            if col != prob_col
        )
        relation = Projection(relation, kept_cols)
    # add a new probability column with name `prob_col` and ones everywhere
    cst_one_probability = Constant[float](
        1.0, auto_infer_type=False, verify_type=False
    )
    relation = ConcatenateConstantColumn(
        relation, str2columnstr_constant(prob_col), cst_one_probability
    )
    relation = RelationalAlgebraSolver().walk(relation)
    return ProvenanceAlgebraSet(relation.value, prob_col)



import typing
from operator import floordiv

from ...datalog import DatalogProgram
from ...datalog.basic_representation import UnionOfConjunctiveQueries
from ...exceptions import ForbiddenDisjunctionError, ForbiddenExpressionError
from ...expression_pattern_matching import add_match
from ...expression_walker import ExpressionWalker, PatternWalker
from ...expressions import Constant, FunctionApplication, Symbol
from ...logic import Implication, Union
from ...type_system import get_generic_type
from ..exceptions import (
    ForbiddenConditionalQueryNoProb,
    MalformedProbabilisticTupleError,
    UnsupportedProbabilisticQueryError,
)
from ..expression_processing import (
    add_to_union,
    build_probabilistic_fact_set,
    check_probabilistic_choice_set_probabilities_sum_to_one,
    get_within_language_prob_query_prob_term,
    group_probabilistic_facts_by_pred_symb,
    is_probabilistic_fact,
    is_within_language_prob_query,
    union_contains_probabilistic_facts,
)
from ..expressions import PROB, Condition, ProbabilisticQuery


def is_within_language_prob_query_wannabe(expression):
    return (
        isinstance(expression, FunctionApplication)
        and get_generic_type(type(expression)) is FunctionApplication
        and expression.functor == PROB
    )


class TranslateProbabilisticQueryMixin(PatternWalker):
    @add_match(
        Implication(
            ..., FunctionApplication(Constant[typing.Any](floordiv), ...)
        )
    )
    def conditional_query(self, implication):
        new_implication = Implication(
            implication.consequent, Condition(*implication.antecedent.args)
        )
        if not (
            any(
                isinstance(arg, ProbabilisticQuery)
                or is_within_language_prob_query_wannabe(arg)
                for arg in new_implication.consequent.args
            )
        ):
            raise ForbiddenExpressionError(
                "Missing probabilistic term in consequent's head"
            )
        if (
            sum(
                isinstance(arg, ProbabilisticQuery)
                for arg in new_implication.consequent.args
            )
            > 1
        ):
            raise ForbiddenExpressionError(
                "Can only contain one probabilistic term in consequent's head"
            )
        return self.walk(new_implication)

    @add_match(
        Implication,
        lambda implication: any(
            is_within_language_prob_query_wannabe(arg)
            for arg in implication.consequent.args
        ),
    )
    def within_language_prob_query(self, implication):
        csqt_args = tuple()
        for arg in implication.consequent.args:
            if is_within_language_prob_query_wannabe(arg):
                arg = ProbabilisticQuery(*arg.unapply())
            csqt_args += (arg,)
        consequent = implication.consequent.functor(*csqt_args)
        return self.walk(Implication(consequent, implication.antecedent))


class CPLogicMixin(PatternWalker):
    """
    Datalog extended with probabilistic facts semantics from ProbLog.

    It adds a probabilistic database which is a set of probabilistic facts.

    Probabilistic facts are stored in the symbol table of the program such that
    the key in the symbol table is the symbol of the predicate of the
    probabilsitic fact and the value is the probabilistic fact itself.
    """

    pfact_pred_symb_set_symb = Symbol("__pfact_pred_symb_set_symb__")
    pchoice_pred_symb_set_symb = Symbol("__pchoice_pred_symb_set_symb__")
    protected_keywords = {"PROB"}

    @property
    def predicate_symbols(self):
        return (
            set(self.intensional_database())
            | set(self.extensional_database())
            | set(self.pfact_pred_symbs)
            | set(self.pchoice_pred_symbs)
        )

    @property
    def probabilistic_predicate_symbols(self):
        return (
            self.pfact_pred_symbs
            | self.pchoice_pred_symbs
            | set(self.within_language_succ_queries())
        )

    @property
    def pfact_pred_symbs(self):
        return self._get_pred_symbs(self.pfact_pred_symb_set_symb)

    @property
    def pchoice_pred_symbs(self):
        return self._get_pred_symbs(self.pchoice_pred_symb_set_symb)

    def within_language_succ_queries(self):
        return {
            pred_symb: union.formulas[0]
            for pred_symb, union in self.intensional_database().items()
            if len(union.formulas) == 1
            and is_within_language_prob_query(union.formulas[0])
        }

    def probabilistic_facts(self):
        """Return probabilistic facts of the symbol table."""
        return {
            k: v
            for k, v in self.symbol_table.items()
            if k in self.pfact_pred_symbs
        }

    def probabilistic_choices(self):
        """Return probabilistic choices of the symbol table."""
        return {
            k: v
            for k, v in self.symbol_table.items()
            if k in self.pchoice_pred_symbs
        }

    def _get_pred_symbs(self, set_symb):
        return self.symbol_table.get(
            set_symb, Constant[typing.AbstractSet](set())
        ).value

    def extensional_database(self):
        exclude = (
            self.protected_keywords
            | self.pfact_pred_symbs
            | self.pchoice_pred_symbs
        )
        ret = self.symbol_table.symbols_by_type(typing.AbstractSet)
        for keyword in exclude:
            if keyword in ret:
                del ret[keyword]
        return ret

    def add_probabilistic_facts_from_tuples(self, symbol, iterable):
        """
        Add probabilistic facts from tuples whose first element
        contains the probability label attached to that tuple.

        Examples
        --------
        The following

        >>> P = Symbol('P')
        >>> tuples = {(0.9, 'a'), (0.8, 'b')}
        >>> program.add_probabilistic_facts_from_tuples(P, tuples)

        adds the probabilistic facts

            P(a) : 0.9  <-  T
            P(b) : 0.8  <-  T

        to the program.

        """
        self._register_prob_pred_symb_set_symb(
            symbol, self.pfact_pred_symb_set_symb
        )
        type_, iterable = self.infer_iterable_type(iterable)
        self._check_iterable_prob_type(type_)
        constant = Constant[typing.AbstractSet[type_]](
            self.new_set(iterable), auto_infer_type=False, verify_type=False
        )
        symbol = symbol.cast(constant.type)
        self.symbol_table[symbol] = constant

    def add_probabilistic_choice_from_tuples(self, symbol, iterable):
        """
        Add a probabilistic choice from a predicate symbol and a set of
        tuples where the first element is the probability label
        attached to a head predicate and the other elements are the
        constant terms of the head predicate.

        Examples
        --------
        The following

        >>> P = Symbol('P')
        >>> tuples = {(0.2, 'a'), (0.8, 'b')}
        >>> program.add_probabilistic_choice_from_tuples(P, tuples)

        adds the probabilistic choice

            P(a) : 0.2  v  P(b) : 0.8  <-  T

        to the program.

        """
        self._register_prob_pred_symb_set_symb(
            symbol, self.pchoice_pred_symb_set_symb
        )
        type_, iterable = self.infer_iterable_type(iterable)
        self._check_iterable_prob_type(type_)
        if symbol in self.symbol_table:
            raise ForbiddenDisjunctionError(
                "Cannot define multiple probabilistic choices with the same "
                f"predicate symbol. Predicate symbol was: {symbol}"
            )
        ra_set = Constant[typing.AbstractSet[type_]](
            self.new_set(iterable), auto_infer_type=False, verify_type=False
        )
        check_probabilistic_choice_set_probabilities_sum_to_one(ra_set)
        self.symbol_table[symbol] = ra_set

    @staticmethod
    def _check_iterable_prob_type(iterable_type):
        if not (
            issubclass(iterable_type.__origin__, typing.Tuple)
            and iterable_type.__args__[0] is float
        ):
            raise MalformedProbabilisticTupleError(
                "Expected tuples to have a probability as their first element"
            )

    @add_match(Union, union_contains_probabilistic_facts)
    def union_with_probabilistic_facts(self, code):
        pfacts, other_expressions = group_probabilistic_facts_by_pred_symb(
            code
        )
        for pred_symb, pfacts in pfacts.items():
            self._register_prob_pred_symb_set_symb(
                pred_symb, self.pfact_pred_symb_set_symb
            )
            if len(pfacts) > 1:
                self.symbol_table[pred_symb] = build_probabilistic_fact_set(
                    pred_symb, pfacts
                )
            else:
                self.walk(list(pfacts)[0])
        self.walk(Union(other_expressions))

    def _register_prob_pred_symb_set_symb(self, pred_symb, set_symb):
        if set_symb.name not in self.protected_keywords:
            self.protected_keywords.add(set_symb.name)
        if set_symb not in self.symbol_table:
            self.symbol_table[set_symb] = Constant[typing.AbstractSet](set())
        self.symbol_table[set_symb] = Constant[typing.AbstractSet](
            self.symbol_table[set_symb].value | {pred_symb}
        )

    @add_match(Implication, is_probabilistic_fact)
    def probabilistic_fact(self, expression):
        pred_symb = expression.consequent.body.functor
        if pred_symb not in self.symbol_table:
            self.symbol_table[pred_symb] = Union(tuple())
        self.symbol_table[pred_symb] = add_to_union(
            self.symbol_table[pred_symb], [expression]
        )
        return expression

    @add_match(Implication(..., Condition), is_within_language_prob_query)
    def within_language_marg_query(self, implication):
        self._validate_within_language_marg_query(implication)
        pred_symb = implication.consequent.functor.cast(
            UnionOfConjunctiveQueries
        )
        if pred_symb in self.symbol_table:
            raise ForbiddenDisjunctionError(
                "Disjunctive within-language probabilistic queries "
                "are not allowed"
            )
        self.symbol_table[pred_symb] = Union((implication,))
        return implication

    @add_match(Implication(..., Condition))
    def marg_implication(self, implication):
        raise ForbiddenConditionalQueryNoProb(
            "Conditional queries which don't have PROB in "
            "the head (within-language probabilistic) "
            "are forbidden"
        )

    @add_match(Implication, is_within_language_prob_query)
    def within_language_succ_query(self, implication):
        self._validate_within_language_succ_query(implication)
        pred_symb = implication.consequent.functor.cast(
            UnionOfConjunctiveQueries
        )
        if pred_symb in self.symbol_table:
            raise ForbiddenDisjunctionError(
                "Disjunctive within-language probabilistic queries "
                "are not allowed"
            )
        self.symbol_table[pred_symb] = Union((implication,))
        return implication

    @staticmethod
    def _validate_within_language_succ_query(implication):
        csqt_vars = set(
            arg
            for arg in implication.consequent.args
            if isinstance(arg, Symbol)
        )
        prob_term = get_within_language_prob_query_prob_term(implication)
        if not prob_term.args:
            raise UnsupportedProbabilisticQueryError(
                "Probabilistic boolean queries are not currently supported"
            )
        if not all(isinstance(arg, Symbol) for arg in prob_term.args):
            bad_vars = (
                repr(arg)
                for arg in prob_term.args
                if not isinstance(arg, Symbol)
            )
            raise ForbiddenExpressionError(
                "All terms in PROB(...) should be variables. "
                "Found these terms: {}".format(", ".join(bad_vars))
            )
        prob_vars = set(prob_term.args)
        if csqt_vars != prob_vars:
            raise ForbiddenExpressionError(
                "Variables of the set-based query and variables in the "
                "PROB(...) term should be the same variables"
            )

    @staticmethod
    def _validate_within_language_marg_query(implication):
        csqt_vars = set(
            arg
            for arg in implication.consequent.args
            if isinstance(arg, Symbol)
        )
        prob_term = get_within_language_prob_query_prob_term(implication)
        if not prob_term.args:
            raise UnsupportedProbabilisticQueryError(
                "Probabilistic boolean queries are not currently supported"
            )
        if not all(isinstance(arg, Symbol) for arg in prob_term.args):
            bad_vars = (
                repr(arg)
                for arg in prob_term.args
                if not isinstance(arg, Symbol)
            )
            raise ForbiddenExpressionError(
                "All terms in PROB(...) should be variables. "
                "Found these terms: {}".format(", ".join(bad_vars))
            )
        prob_vars = set(prob_term.args)
        if csqt_vars != prob_vars:
            raise ForbiddenExpressionError(
                "Variables of the set-based query and variables in the "
                "PROB(...) term should be the same variables"
            )


class CPLogicProgram(CPLogicMixin, DatalogProgram, ExpressionWalker):
    pass



from . import (
    Symbol,
    Constant,
    FunctionApplication,
    Implication,
    Union,
    Conjunction,
    Disjunction,
    Negation,
    UniversalPredicate,
    ExistentialPredicate,
    Quantifier,
)
from ..logic.expression_processing import ExtractFreeVariablesWalker
from ..expression_walker import (
    ExpressionWalker,
    add_match,
    PatternWalker,
    ChainedWalker,
    ReplaceSymbolWalker,
)


class LogicExpressionWalker(PatternWalker):
    @add_match(Quantifier)
    def walk_quantifier(self, expression):
        return expression.apply(
            self.walk(expression.head), self.walk(expression.body)
        )

    @add_match(Constant)
    def walk_constant(self, expression):
        return expression

    @add_match(Symbol)
    def walk_symbol(self, expression):
        return expression

    @add_match(Negation)
    def walk_negation(self, expression):
        return expression.apply(self.walk(expression.formula))

    @add_match(FunctionApplication)
    def walk_function(self, expression):
        return expression.apply(
            expression.functor, tuple(map(self.walk, expression.args))
        )

    @add_match(Union)
    def walk_union(self, expression):
        return expression.apply(self.walk(expression.formulas))

    @add_match(Disjunction)
    def walk_disjunction(self, expression):
        return expression.apply(map(self.walk, expression.formulas))

    @add_match(Conjunction)
    def walk_conjunction(self, expression):
        return expression.apply(map(self.walk, expression.formulas))


class EliminateImplications(LogicExpressionWalker):
    """
    Removes the implication ocurrences of an expression.
    """

    @add_match(Implication(..., ...))
    def remove_implication(self, implication):
        c = self.walk(implication.consequent)
        a = self.walk(implication.antecedent)
        return self.walk(Disjunction((c, Negation(a))))


class MoveNegationsToAtoms(LogicExpressionWalker):
    """
    Moves the negations the furthest possible to the atoms.
    Assumes that there are no implications in the expression.
    """

    @add_match(Negation(UniversalPredicate(..., ...)))
    def negated_universal(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        p = self.walk(Negation(quantifier.body))
        return self.walk(ExistentialPredicate(x, p))

    @add_match(Negation(ExistentialPredicate(..., ...)))
    def negated_existential(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        p = self.walk(Negation(quantifier.body))
        return self.walk(UniversalPredicate(x, p))

    @add_match(Negation(Conjunction(...)))
    def negated_conjunction(self, negation):
        conj = negation.formula
        formulas = map(lambda e: self.walk(Negation(e)), conj.formulas)
        return self.walk(Disjunction(tuple(formulas)))

    @add_match(Negation(Disjunction(...)))
    def negated_disjunction(self, negation):
        disj = negation.formula
        formulas = map(lambda e: self.walk(Negation(e)), disj.formulas)
        return self.walk(Conjunction(tuple(formulas)))

    @add_match(Negation(Negation(...)))
    def negated_negation(self, negation):
        return self.walk(negation.formula.formula)


class MoveQuantifiersUp(LogicExpressionWalker):
    """
    Moves the quantifiers up in order to format the expression
    in prenex normal form. Assumes the expression contains no implications
    and the variables of the quantifiers are not repeated.
    """

    @add_match(Negation(UniversalPredicate(..., ...)))
    def negated_universal(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        return self.walk(ExistentialPredicate(x, Negation(quantifier.body)))

    @add_match(Negation(ExistentialPredicate(..., ...)))
    def negated_existential(self, negation):
        quantifier = negation.formula
        x = quantifier.head
        return self.walk(UniversalPredicate(x, Negation(quantifier.body)))

    @add_match(
        Disjunction,
        lambda exp: any(isinstance(f, Quantifier) for f in exp.formulas),
    )
    def disjunction_with_quantifiers(self, expression):
        expression = self.walk_disjunction(expression)
        quantifiers = []
        formulas = []
        for f in expression.formulas:
            if isinstance(f, Quantifier):
                quantifiers.append(f)
                formulas.append(f.body)
            else:
                formulas.append(f)
        exp = Disjunction(tuple(formulas))
        for q in reversed(quantifiers):
            exp = q.apply(q.head, exp)
        return self.walk(exp)

    @add_match(
        Conjunction,
        lambda exp: any(isinstance(f, Quantifier) for f in exp.formulas),
    )
    def conjunction_with_quantifiers(self, expression):
        expression = self.walk_conjunction(expression)
        quantifiers = []
        formulas = []
        for f in expression.formulas:
            if isinstance(f, Quantifier):
                quantifiers.append(f)
                formulas.append(f.body)
            else:
                formulas.append(f)
        exp = Conjunction(tuple(formulas))
        for q in reversed(quantifiers):
            exp = q.apply(q.head, exp)
        return self.walk(exp)


class DesambiguateQuantifiedVariables(LogicExpressionWalker):
    """
    Replaces each quantified variale to a fresh one.
    """

    @add_match(Implication)
    def implication(self, expression):
        fs = self.walk(expression.consequent), self.walk(expression.antecedent)
        self._desambiguate(fs)
        return expression.apply(*fs)

    @add_match(Union)
    def union(self, expression):
        fs = self.walk(expression.formulas)
        self._desambiguate(fs)
        return expression.apply(fs)

    @add_match(Disjunction)
    def disjunction(self, expression):
        fs = self.walk(expression.formulas)
        self._desambiguate(fs)
        return expression.apply(fs)

    @add_match(Conjunction)
    def conjunction(self, expression):
        fs = self.walk(expression.formulas)
        self._desambiguate(fs)
        return expression.apply(fs)

    @add_match(Negation)
    def negation(self, expression):
        return expression.apply(self.walk(expression.formula))

    @add_match(Quantifier)
    def quantifier(self, expression):
        expression.body = self.walk(expression.body)
        uq = UsedQuantifiers().walk(expression.body)
        for q in uq:
            if q.head == expression.head:
                self.rename_quantifier(q)
        return expression

    def rename_quantifier(self, q):
        ns = Symbol.fresh()
        q.body = ReplaceFreeSymbolWalker({q.head: ns}).walk(q.body)
        q.head = ns

    def _desambiguate(self, l):
        used_variables = set()
        for f in l:
            uq = UsedQuantifiers().walk(f)
            for q in self._conflicted_quantifiers(used_variables, uq):
                self.rename_quantifier(q)
            used_variables |= self._bound_variables(uq)

    def _conflicted_quantifiers(self, used_variables, quantifiers):
        bv = self._bound_variables(quantifiers)
        repeated = bv & used_variables
        return [q for q in quantifiers if q.head in repeated]

    def _bound_variables(self, quantifiers):
        return set(map(lambda q: q.head, quantifiers))


class ReplaceFreeSymbolWalker(ReplaceSymbolWalker):
    @add_match(Quantifier)
    def stop_if_bound(self, expression):
        s = expression.head
        r = self.symbol_replacements.pop(s, None)
        expression.body = self.walk(expression.body)
        if r:
            self.symbol_replacements[s] = r
        return expression


class UsedQuantifiers(PatternWalker):
    @add_match(FunctionApplication)
    def function(self, exp):
        return set()

    @add_match(Symbol)
    def symbol(self, exp):
        return set()

    @add_match(Negation)
    def negation(self, exp):
        return self.walk(exp.formula)

    @add_match(Disjunction)
    def disjunction(self, exp):
        return set.union(*self.walk(exp.formulas))

    @add_match(Implication)
    def implication(self, exp):
        return set.union(self.walk(exp.antecedent), self.walk(exp.consequent))

    @add_match(Conjunction)
    def conjunction(self, exp):
        return set.union(*self.walk(exp.formulas))

    @add_match(Quantifier)
    def quantifier(self, exp):
        return {exp} | self.walk(exp.body)


class DistributeDisjunctions(LogicExpressionWalker):
    @add_match(Disjunction, lambda e: len(e.formulas) > 2)
    def split(self, expression):
        head, *rest = expression.formulas
        rest = self.walk(Disjunction(tuple(rest)))
        new_exp = Disjunction((head, rest))
        return self.walk(new_exp)

    @add_match(Disjunction((..., Conjunction)))
    def rotate(self, expression):
        q, c = expression.formulas
        return self.walk(
            Conjunction(tuple(map(lambda p: Disjunction((q, p)), c.formulas)))
        )

    @add_match(Disjunction((Conjunction, ...)))
    def distribute(self, expression):
        c, q = expression.formulas
        return self.walk(
            Conjunction(tuple(map(lambda p: Disjunction((p, q)), c.formulas)))
        )


class CollapseDisjunctionsMixin(PatternWalker):
    @add_match(
        Disjunction,
        lambda e: any(isinstance(f, Disjunction) for f in e.formulas),
    )
    def disjunction(self, e):
        new_arg = []
        for f in map(self.walk, e.formulas):
            if isinstance(f, Disjunction):
                new_arg.extend(f.formulas)
            else:
                new_arg.append(f)
        return self.walk(Disjunction(tuple(new_arg)))


class CollapseDisjunctions(CollapseDisjunctionsMixin, ExpressionWalker):
    pass


class CollapseConjunctionsMixin(PatternWalker):
    @add_match(
        Conjunction,
        lambda e: any(isinstance(f, Conjunction) for f in e.formulas),
    )
    def conjunction(self, e):
        new_arg = []
        for f in map(self.walk, e.formulas):
            if isinstance(f, Conjunction):
                new_arg.extend(f.formulas)
            else:
                new_arg.append(f)
        return self.walk(Conjunction(tuple(new_arg)))


class CollapseConjunctions(CollapseConjunctionsMixin, ExpressionWalker):
    pass


class RemoveUniversalPredicates(LogicExpressionWalker):
    """
    Changes the universal predicates by equivalent expressions
    using existential quantifiers.
    """

    @add_match(UniversalPredicate)
    def universal(self, expression):
        return self.walk(
            Negation(
                ExistentialPredicate(
                    expression.head, Negation(self.walk(expression.body))
                )
            )
        )


def convert_to_pnf_with_cnf_matrix(expression):
    walker = ChainedWalker(
        EliminateImplications,
        MoveNegationsToAtoms,
        DesambiguateQuantifiedVariables,
        MoveQuantifiersUp,
        DistributeDisjunctions,
        CollapseDisjunctions,
        CollapseConjunctions,
    )
    return walker.walk(expression)


class ExtractFOLFreeVariables(ExtractFreeVariablesWalker):
    @add_match(Implication)
    def extract_variables_s(self, exp):
        return self.walk(exp.consequent) | self.walk(exp.antecedent)


class DistributeUniversalQuantifiers(PatternWalker):
    @add_match(UniversalPredicate(..., Conjunction))
    def distribute_universal_quantifier(self, uq):
        return self.walk(
            Conjunction(
                tuple(map(self._apply_quantifier(uq.head), uq.body.formulas))
            )
        )

    def _apply_quantifier(self, var):
        def foo(exp):
            fv = ExtractFOLFreeVariables().walk(exp)
            if var in fv:
                exp = UniversalPredicate(var, exp)
            return exp

        return foo


class DistributeImplicationsWithConjunctiveHeads(PatternWalker):
    @add_match(Implication(Conjunction, ...))
    def distribute_implication_with_conjunctive_head(self, impl):
        return self.walk(
            Conjunction(
                tuple(
                    Implication(h, impl.antecedent)
                    for h in impl.consequent.formulas
                )
            )
        )



import operator

from ...datalog.basic_representation import DatalogProgram
from ...expression_walker import ExpressionWalker
from ...expressions import Constant, Symbol
from ...logic import Conjunction, Implication, Union
from ..expression_processing import (
    CollapseConjunctiveAntecedents,
    RemoveDuplicatedAntecedentPredicates,
    TranslateToLogic,
    UnifyVariableEqualitiesMixin,
)

P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")
a = Constant("a")
b = Constant("b")
c = Constant("c")

EQ = Constant(operator.eq)


class DatalogWithVariableEqualityPropagation(
    TranslateToLogic,
    CollapseConjunctiveAntecedents,
    UnifyVariableEqualitiesMixin,
    RemoveDuplicatedAntecedentPredicates,
    DatalogProgram,
    ExpressionWalker,
):
    pass


def test_propagation_to_one_conjunct():
    rule = Implication(R(x), Conjunction((P(x, y), EQ(y, a))))
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R in program.intensional_database()
    expected = Union((Implication(R(x), P(x, a)),))
    result = program.intensional_database()[R]
    assert result == expected


def test_propagation_to_two_conjuncts():
    rule = Implication(R(x, y), Conjunction((P(x, y), EQ(y, a), Q(y, y))))
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R in program.intensional_database()
    expected = Union((Implication(R(x, a), Conjunction((P(x, a), Q(a, a)))),))
    assert expected == program.intensional_database()[R]


def test_single_equality_antecedent():
    rule = Implication(R(x), Conjunction((EQ(x, a),)))
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R not in program.intensional_database()
    assert R in program.extensional_database()


def test_between_vars_equality_propagation():
    rule = Implication(R(x, y), Conjunction((P(x, y), EQ(y, x), Q(y, y))))
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R in program.intensional_database()
    expecteds = [
        Union((Implication(R(x, x), Conjunction((P(x, x), Q(x, x)))),)),
        Union((Implication(R(y, y), Conjunction((P(y, y), Q(y, y)))),)),
    ]
    result = program.intensional_database()[R]
    assert any(hash(expected) == hash(result) for expected in expecteds)


def test_multiple_between_vars_equalities():
    rule = Implication(
        R(x, y, z), Conjunction((P(z, x), EQ(y, z), EQ(y, x), Q(y, y)))
    )
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R in program.intensional_database()
    expecteds = [
        Union((Implication(R(v, v, v), Conjunction((P(v, v), Q(v, v)))),))
        for v in (x, y, z)
    ]
    result = program.intensional_database()[R]
    assert any(hash(expected) == hash(result) for expected in expecteds)


def test_mix_between_var_eqs_var_to_const_eq():
    rule = Implication(
        R(x, y, z),
        Conjunction((P(z, x), EQ(y, z), EQ(y, a), Q(y, y), EQ(b, x))),
    )
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    assert R in program.intensional_database()
    expected = Union(
        (
            Implication(
                R(b, a, a),
                Conjunction((P(a, b), Q(a, a))),
            ),
        )
    )
    result = program.intensional_database()[R]
    assert hash(result) == hash(expected)


def test_collapsable_conjunction():
    rule = Implication(
        R(x, y),
        Conjunction(
            (
                Conjunction((P(z), Q(y, x), EQ(y, z))),
                Conjunction((Q(z, y), P(z), P(y), EQ(z, a))),
                Z(b),
                Z(z),
            )
        ),
    )
    program = DatalogWithVariableEqualityPropagation()
    program.walk(rule)
    expected = Union(
        (
            Implication(
                R(x, a),
                Conjunction((P(a), Q(a, x), Q(a, a), Z(b), Z(a))),
            ),
        )
    )
    result = program.intensional_database()[R]
    assert isinstance(result, Union)
    assert len(result.formulas) == 1
    assert isinstance(result.formulas[0], Implication)
    assert result.formulas[0].consequent == expected.formulas[0].consequent
    assert set(result.formulas[0].antecedent.formulas) == set(
        expected.formulas[0].antecedent.formulas
    )



"""
Implentation of probabilistic query resolution for
hierarchical queries [^1]. Using this we apply the small dichotomy
theorem [^1, ^2]:
Let Q be a conjunctive query without self-joins or a non-repeating relational
algebra expression. Then:
* If Q is hierarchical, then P(Q) is in polynomial time, and can be computed
using only the lifted inference rules for join, negation, union, and
existential quantifier.

* If Q is not hierarchical, then P(Q) is #P-hard in the size of the database.

[^1]: Robert Fink and Dan Olteanu. A dichotomy for non-repeating queries with
negation in probabilistic databases. In Proceedings of the 33rd ACM
SIGMOD-SIGACT-SIGART Symposium on Principles of Database Systems, PODS ’14,
pages 144–155, New York, NY, USA, 2014. ACM.

[^2]: Nilesh N. Dalvi and Dan Suciu. Efficient query evaluation on
probabilistic databases. VLDB J., 16(4):523–544, 2007.
"""

import logging
import typing
from collections import defaultdict

from ..datalog.expression_processing import enforce_conjunction, flatten_query
from ..datalog.translate_to_named_ra import TranslateToNamedRA
from ..expression_walker import ExpressionWalker, add_match
from ..expressions import Constant, Symbol
from ..logic import FALSE, Conjunction, Implication
from ..logic.expression_processing import (
    extract_logic_atoms,
    extract_logic_free_variables,
    extract_logic_predicates,
)
from ..relational_algebra import (
    ColumnInt,
    ColumnStr,
    EliminateTrivialProjections,
    ExtendedProjection,
    ExtendedProjectionListMember,
    NameColumns,
    NamedRelationalAlgebraFrozenSet,
    Projection,
    RelationalAlgebraPushInSelections,
    RelationalAlgebraStringExpression,
    str2columnstr_constant,
)
from ..relational_algebra_provenance import (
    NaturalJoinInverse,
    ProvenanceAlgebraSet,
    RelationalAlgebraProvenanceCountingSolver,
    RelationalAlgebraProvenanceExpressionSemringSolver,
)
from ..utils import log_performance
from ..utils.orderedset import OrderedSet
from .exceptions import NotHierarchicalQueryException
from .expression_processing import (
    lift_optimization_for_choice_predicates,
    project_on_query_head,
)
from .probabilistic_ra_utils import (
    DeterministicFactSet,
    ProbabilisticChoiceSet,
    ProbabilisticFactSet,
    generate_probabilistic_symbol_table_for_query,
)
from .shattering import shatter_easy_probfacts

LOG = logging.getLogger(__name__)


def is_hierarchical_without_self_joins(query):
    """
    Let Q be first-order formula. For each variable x denote at(x) the
    set of atoms that contain the variable x. We say that Q is hierarchical
    if forall x, y one of the following holds:
    at(x) ⊆ at(y) or at(x) ⊇ at(y) or at(x) ∩ at(y) = ∅.
    """

    has_self_joins, atom_set = extract_atom_sets_and_detect_self_joins(query)

    if has_self_joins:
        return False

    variables = list(atom_set)
    for i, v in enumerate(variables):
        at_v = atom_set[v]
        for v2 in variables[i + 1 :]:
            at_v2 = atom_set[v2]
            if not (at_v <= at_v2 or at_v2 <= at_v or at_v.isdisjoint(at_v2)):
                LOG.info(
                    "Not hierarchical on variables %s %s", v.name, v2.name
                )
                return False

    return True


def extract_atom_sets_and_detect_self_joins(query):
    has_self_joins = False
    predicates = extract_logic_atoms(query)
    seen_predicate_functor = set()
    atom_set = defaultdict(set)
    for predicate in predicates:
        functor = predicate.functor
        if functor in seen_predicate_functor:
            LOG.info("Not hierarchical self join on variables %s", functor)
            has_self_joins = True
        seen_predicate_functor.add(functor)
        for variable in predicate.args:
            if not isinstance(variable, Symbol):
                continue
            atom_set[variable].add(functor)
    return has_self_joins, atom_set


class ProbSemiringSolver(RelationalAlgebraProvenanceExpressionSemringSolver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.translated_probfact_sets = dict()

    @add_match(
        Projection,
        lambda exp: (
            isinstance(
                exp.relation,
                (
                    DeterministicFactSet,
                    ProbabilisticFactSet,
                    ProbabilisticChoiceSet,
                ),
            )
        ),
    )
    def eliminate_superfluous_projection(self, expression):
        return self.walk(expression.relation)

    @add_match(DeterministicFactSet(Symbol))
    def deterministic_fact_set(self, deterministic_set):
        relation_symbol = deterministic_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        relation = self.walk(relation_symbol)
        named_columns = tuple(
            str2columnstr_constant(f"col_{i}") for i in relation.value.columns
        )
        projection_list = [
            ExtendedProjectionListMember(
                Constant[RelationalAlgebraStringExpression](
                    RelationalAlgebraStringExpression(c.value),
                    verify_type=False,
                ),
                c,
            )
            for c in named_columns
        ]

        prov_column = ColumnStr(Symbol.fresh().name)
        provenance_set = self.walk(
            ExtendedProjection(
                NameColumns(relation, named_columns),
                tuple(projection_list)
                + (
                    ExtendedProjectionListMember(
                        Constant[float](1.0),
                        str2columnstr_constant(prov_column),
                    ),
                ),
            )
        )

        self.translated_probfact_sets[relation_symbol] = ProvenanceAlgebraSet(
            provenance_set.value, prov_column
        )
        return self.translated_probfact_sets[relation_symbol]

    @add_match(ProbabilisticFactSet(Symbol, ...))
    def probabilistic_fact_set(self, prob_fact_set):
        relation_symbol = prob_fact_set.relation
        if relation_symbol in self.translated_probfact_sets:
            return self.translated_probfact_sets[relation_symbol]

        relation = self.walk(relation_symbol)
        named_columns = tuple(
            str2columnstr_constant(f"col_{i}") for i in relation.value.columns
        )
        relation = NameColumns(relation, named_columns)
        relation = self.walk(relation)
        rap_column = ColumnStr(
            relation.value.columns[prob_fact_set.probability_column.value]
        )

        self.translated_probfact_sets[relation_symbol] = ProvenanceAlgebraSet(
            relation.value, rap_column
        )
        return self.translated_probfact_sets[relation_symbol]

    @add_match(ProbabilisticChoiceSet(Symbol, ...))
    def probabilistic_choice_set(self, prob_choice_set):
        return self.probabilistic_fact_set(prob_choice_set)

    @add_match(ProbabilisticFactSet)
    def probabilistic_fact_set_invalid(self, prob_fact_set):
        raise NotImplementedError()

    @add_match(ExtendedProjection(ProvenanceAlgebraSet, ...))
    def extended_projection(self, proj_op):
        provset = self.walk(proj_op.relation)
        self._check_prov_col_not_in_proj_list(provset, proj_op.projection_list)
        self._check_all_non_prov_cols_in_proj_list(
            provset, proj_op.projection_list
        )
        relation = Constant[typing.AbstractSet](provset.relations)
        prov_col = str2columnstr_constant(provset.provenance_column)
        new_prov_col = str2columnstr_constant(Symbol.fresh().name)
        proj_list_with_prov_col = proj_op.projection_list + (
            ExtendedProjectionListMember(prov_col, new_prov_col),
        )
        ra_op = ExtendedProjection(relation, proj_list_with_prov_col)
        new_relation = self.walk(ra_op)
        new_provset = ProvenanceAlgebraSet(
            new_relation.value, new_prov_col.value
        )
        return new_provset

    @staticmethod
    def _check_prov_col_not_in_proj_list(provset, proj_list):
        if any(
            member.dst_column.value == provset.provenance_column
            for member in proj_list
        ):
            raise ValueError(
                "Cannot project on provenance column: "
                f"{provset.provenance_column}"
            )

    @staticmethod
    def _check_all_non_prov_cols_in_proj_list(provset, proj_list):
        non_prov_cols = set(provset.non_provenance_columns)
        found_cols = set(
            member.dst_column.value
            for member in proj_list
            if member.dst_column.value in non_prov_cols
            and member.fun_exp == member.dst_column
        )
        if non_prov_cols.symmetric_difference(found_cols):
            raise ValueError(
                "All non-provenance columns must be part of the extended "
                "projection as {c: c} projection list member."
            )


class RAQueryOptimiser(
    EliminateTrivialProjections,
    RelationalAlgebraPushInSelections,
    ExpressionWalker,
):
    pass


def solve_succ_query(query, cpl_program):
    """
    Solve a SUCC query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        SUCC query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """
    with log_performance(
        LOG,
        "Preparing query %s",
        init_args=(query.consequent.functor.name,),
    ):
        flat_query_body = flatten_query(query.antecedent, cpl_program)

    if flat_query_body == FALSE or (
        isinstance(flat_query_body, Conjunction)
        and any(conjunct == FALSE for conjunct in flat_query_body.formulas)
    ):
        return ProvenanceAlgebraSet(
            NamedRelationalAlgebraFrozenSet(("_p_",)),
            ColumnStr("_p_"),
        )

    with log_performance(LOG, "Translation and lifted optimisation"):
        flat_query_body = enforce_conjunction(
            lift_optimization_for_choice_predicates(
                flat_query_body, cpl_program
            )
        )
        flat_query = Implication(query.consequent, flat_query_body)
        symbol_table = generate_probabilistic_symbol_table_for_query(
            cpl_program, flat_query_body
        )
        shattered_query = shatter_easy_probfacts(flat_query, symbol_table)
        prob_pred_symbs = (
            cpl_program.pfact_pred_symbs | cpl_program.pchoice_pred_symbs
        )
        # note: this assumes that the shattering process does not change the
        # order of the antecedent's conjuncts
        shattered_query_probabilistic_body = Conjunction(
            tuple(
                shattered_conjunct
                for shattered_conjunct, flat_conjunct in zip(
                    shattered_query.antecedent.formulas,
                    flat_query.antecedent.formulas,
                )
                if extract_logic_atoms(flat_conjunct)[0].functor
                in prob_pred_symbs
            )
        )
        if not is_hierarchical_without_self_joins(
            shattered_query_probabilistic_body
        ):
            LOG.info(
                "Query with conjunctions %s not hierarchical",
                shattered_query_probabilistic_body.formulas,
            )
            raise NotHierarchicalQueryException(
                "Query not hierarchical, algorithm can't be applied"
            )

        ra_query = TranslateToNamedRA().walk(shattered_query.antecedent)
        # finally project on the initial query's head variables
        proj_cols = tuple(
            OrderedSet(
                str2columnstr_constant(arg.name)
                for arg in query.consequent.args
            )
        )
        ra_query = Projection(ra_query, proj_cols)
        ra_query = RAQueryOptimiser().walk(ra_query)

    with log_performance(LOG, "Run RAP query"):
        solver = ProbSemiringSolver(symbol_table)
        prob_set_result = solver.walk(ra_query)

    return prob_set_result


def solve_marg_query(rule, cpl):
    """
    Solve a MARG query on a CP-Logic program.

    Parameters
    ----------
    query : Implication
        Consequent must be of type `Condition`.
        MARG query of the form `ans(x) :- P(x)`.
    cpl_program : CPLogicProgram
        CP-Logic program on which the query should be solved.

    Returns
    -------
    ProvenanceAlgebraSet
        Provenance set labelled with probabilities for each tuple in the result
        set.

    """
    res_args = tuple(s for s in rule.consequent.args if isinstance(s, Symbol))

    joint_antecedent = Conjunction(
        tuple(
            extract_logic_predicates(rule.antecedent.conditioned)
            | extract_logic_predicates(rule.antecedent.conditioning)
        )
    )
    joint_logic_variables = (
        extract_logic_free_variables(joint_antecedent) & res_args
    )
    joint_rule = Implication(
        Symbol.fresh()(*joint_logic_variables), joint_antecedent
    )
    joint_provset = solve_succ_query(joint_rule, cpl)

    denominator_antecedent = rule.antecedent.conditioning
    denominator_logic_variables = (
        extract_logic_free_variables(denominator_antecedent) & res_args
    )
    denominator_rule = Implication(
        Symbol.fresh()(*denominator_logic_variables), denominator_antecedent
    )
    denominator_provset = solve_succ_query(denominator_rule, cpl)
    rapcs = RelationalAlgebraProvenanceCountingSolver()
    provset = rapcs.walk(
        Projection(
            NaturalJoinInverse(joint_provset, denominator_provset),
            tuple(str2columnstr_constant(s.name) for s in res_args),
        )
    )
    return provset



'''
Implementation of the type system based on Siek and Vachharajani,
"Gradual Typing with Unification-based Inference", DLS 2008.
'''


import inspect
import operator
import sys
import types
from itertools import islice
from typing import (AbstractSet, Any, Callable, Generic, Iterable, Mapping,
                    Sequence, Set, Text, Tuple, TypeVar)

import numpy as np
from typing_inspect import (get_origin, is_callable_type, is_generic_type,
                            is_tuple_type, is_typevar, is_union_type)

from ..exceptions import NeuroLangException

NEW_TYPING = sys.version_info[:3] >= (3, 7, 0)


class NeuroLangTypeException(NeuroLangException):
    pass


if sys.version_info < (3, 6, 0):
    raise ImportError("Only python 3.6 and over compatible")
if not NEW_TYPING:
    from typing import _FinalTypingBase

    class _Unknown(_FinalTypingBase, _root=True):
        """Special type indicating an unknown type.

        - Unknown is compatible with every type.
        - Unknown is less informative than all types.
        """

        __slots__ = ()

        def __instancecheck__(self, obj):
            raise TypeError("Unknown cannot be used with isinstance().")

        def __subclasscheck__(self, cls):
            raise TypeError("Unknown cannot be used with issubclass().")

    Unknown = _Unknown(_root=True)
else:
    from typing import _SpecialForm, _Final, _Immutable, _GenericAlias

    Unknown = _SpecialForm(
        'Unknown', doc="""
        Special type indicating an unknown type.

        - Unknown is compatible with every type.
        - Unknown is less informative than all types.
        """
    )

    class Unknown(_Final, _Immutable, _root=True):
        """Special type indicating an unknown type.

        - Unknown is compatible with every type.
        - Unknown is less informative than all types.
        """

        __slots__ = ()

        def __instancecheck__(self, obj):
            raise TypeError("Unknown cannot be used with isinstance().")

        def __subclasscheck__(self, cls):
            raise TypeError("Unknown cannot be used with issubclass().")


type_order = {
    np.int64: (np.int,),
    np.bool_: (bool,),
    bool: (np.bool_,),
    int: (np.int64, float, complex),
    float: (np.float64, complex,),
    Set: (AbstractSet,)
}


def is_consistent(type1, type2):
    if not isinstance(type1, type) and isinstance(type2, type):
        raise ValueError('Both parameters need to be types')

    if type1 is Unknown or type2 is Unknown:
        return True
    elif is_parameterized(type1) and is_parameterized(type2):
        generic1 = get_origin(type1)
        generic2 = get_origin(type2)

        if not is_consistent(generic1, generic2):
            return False

        type_parameters1 = get_args(type1)
        type_parameters2 = get_args(type2)

        return len(type_parameters1) == len(type_parameters2) and all(
            is_consistent(t1, t2)
            for t1, t2 in zip(type_parameters1, type_parameters2)
        )
    else:
        return type1 is type2


def is_leq_informative(left, right):
    if not (is_type(left) and is_type(right)):
        raise ValueError('Both parameters need to be types')
    if (
        get_origin(left) is Generic or
        get_origin(right) is Generic
    ):
        raise ValueError("typing Generic not supported")
    if left is right:
        result = True
    elif left is Unknown:
        result = True
    elif right is Unknown:
        result = False
    elif right is Any:
        result = True
    elif left is Any:
        result = False
    elif is_union_type(right):
        result = is_leq_informative_union(left, right)
    elif is_union_type(left):
        result = False
    elif is_parameterized(right):
        result = is_leq_informative_parameterized_right(left, right)
    elif is_parametrical(right):
        if is_parameterized(left):
            left = get_origin(left)
        result = issubclass(left, right)
    elif is_parametrical(left) or is_parameterized(left):
        result = False
    elif left in type_order and right in type_order[left]:
        result = True
    else:
        result = issubclass(left, right)

    return result


def is_leq_informative_union(left, right):
    type_parameters_right = get_args(right)
    if is_union_type(left):
        type_parameters_left = get_args(left)
        return all(
            any(
                is_leq_informative(l, r)
                for r in type_parameters_right
            )
            for l in type_parameters_left
        )
    else:
        return any(
            is_leq_informative(left, parameter)
            for parameter in get_args(right)
        )


def is_leq_informative_parameterized_right(left, right):
    generic_right = get_origin(right)

    if is_parameterized(left):
        if not is_leq_informative(get_origin(left), generic_right):
            return False

        type_parameters_left = get_args(left)
        type_parameters_right = get_args(right)

        if len(type_parameters_left) != len(type_parameters_right):
            return False

        return all(
            is_leq_informative(t_left, t_right)
            for t_left, t_right in
            zip(type_parameters_left, type_parameters_right)
        )
    elif is_parametrical(left):
        return False
    else:
        return is_leq_informative(left, generic_right)


def is_type(type_):
    return (
        isinstance(type_, type) or
        type_ is Unknown or
        type_ is Any or
        is_parameterized(type_) or
        is_parametrical(type_) or
        is_typevar(type_) or
        is_union_type(type_)

    )


def is_parametrical(type_):
    is_parametrical_generic = any(
        p(type_)
        for p in
        (is_generic_type, is_callable_type, is_tuple_type, is_union_type)
    ) and not getattr(type_, '_is_protocol', False)

    if is_parametrical_generic:
        if NEW_TYPING:
            return getattr(type_, '_special', False) or (
                is_union_type(type_) and not hasattr(type_, '__args__')
            )
        else:
            return type_.__args__ is None
    else:
        return False


def is_parameterized(type_):
    is_parametrical_generic = any(
        p(type_)
        for p in
        (is_generic_type, is_callable_type, is_tuple_type, is_union_type)
    )

    if is_parametrical_generic:
        if NEW_TYPING:
            return not (
                getattr(type_, '_special', False) or (
                    is_union_type(type_) and not hasattr(type_, '__args__')
                ) or
                getattr(type_, '_is_protocol', False)
            )
        else:
            return get_origin(type_) is not type_
    else:
        return False


def unify_types(t1, t2):
    if t1 is Unknown:
        return t2
    elif t2 is Unknown:
        return t1
    elif is_leq_informative(t1, t2):
        return t2
    elif is_leq_informative(t2, t1):
        return t1
    else:
        raise NeuroLangTypeException(
            "The types {} and {} can't be unified".format(
                t1, t2
            )
        )


def infer_type(value, deep=False, recursive_callback=None):
    if recursive_callback is None:
        recursive_callback = infer_type

    if isinstance(value, (types.FunctionType, types.MethodType)):
        result = typing_callable_from_annotated_function(value)
    elif isinstance(value, types.BuiltinFunctionType):
        result = infer_type_builtins(value)
    elif isinstance(value, Tuple):
        inner_types = tuple(
            recursive_callback(v)
            for v in value
        )
        result = Tuple[inner_types]
    elif isinstance(value, Text):
        result = type(value)
    elif isinstance(value, (AbstractSet, Sequence)):
        result = infer_type_iterables(
            value, deep=deep, recursive_callback=recursive_callback
        )
    elif isinstance(value, Mapping):
        result = infer_type_mapping(
            value, deep=deep, recursive_callback=recursive_callback
        )
    else:
        result = type(value)

    return result


def infer_type_iterables(value, deep=True, recursive_callback=infer_type):
    inner_type = Unknown
    it = iter(value)
    if not deep:
        it = islice(it, 1)

    for element in it:
        inner_type = unify_types(
            recursive_callback(element), inner_type
        )
    if isinstance(value, AbstractSet):
        return AbstractSet[inner_type]
    elif isinstance(value, Sequence):
        return Sequence[inner_type]


def infer_type_mapping(value,  deep=True, recursive_callback=infer_type):
    ktype = Unknown
    vtype = Unknown
    it = iter(value.items())
    if not deep:
        it = islice(it, 1)

    for k, v in it:
        ktype = unify_types(recursive_callback(k), ktype)
        vtype = unify_types(recursive_callback(v), vtype)
    return Mapping[ktype, vtype]


def replace_type_variable(type_, type_hint, type_var=None):
    if (
        isinstance(type_hint, TypeVar) and
        type_hint == type_var
    ):
        return type_
    elif is_parameterized(type_hint):
        new_args = replace_type_variable(
            type_, get_args(type_hint), type_var=type_var
        )
        new_args = tuple(new_args)
        origin = get_origin(type_hint)
        return replace_type_variable_fix_python36_37(
            type_hint, origin, new_args
        )
    elif isinstance(type_hint, Iterable):
        return [
            replace_type_variable(type_, arg, type_var=type_var)
            for arg in type_hint
        ]
    else:
        return type_hint


def replace_type_variable_fix_python36_37(type_hint, origin, new_args):
    if NEW_TYPING and isinstance(type_hint, _GenericAlias):
        new_type = type_hint.copy_with(new_args)
    else:
        new_type = origin[new_args]
    return new_type


def typing_callable_from_annotated_function(function):
    """Get typing.Callable type representing the annotated function type."""
    signature = inspect.signature(function)
    parameter_types = [
        v.annotation if v.annotation is not inspect.Parameter.empty
        else Unknown
        for v in signature.parameters.values()
    ]

    if signature.return_annotation is inspect.Parameter.empty:
        return_annotation = Unknown
    else:
        return_annotation = signature.return_annotation
    return Callable[
        parameter_types,
        return_annotation
    ]


_BINARY_OPERATORS = (
        operator.eq,
        operator.ne,
        operator.gt,
        operator.ge,
        operator.lt,
        operator.le,
        operator.and_,
        operator.or_,
        operator.xor,
        operator.contains,
        operator.matmul,
        operator.mul,
        operator.add,
        operator.sub,
        operator.pow,
        operator.lshift,
        operator.rshift,
        operator.truediv,
        operator.ior,
        operator.iand,
)


_UNARY_OPERATORS = (
    operator.neg,
    operator.pos,
    operator.invert,
)


_NARY_OPERATORS = (
    sum,
)

_RELATIVE_OPERATORS = (
    operator.eq,
    operator.ne,
    operator.gt,
    operator.ge,
    operator.lt,
    operator.le,
    operator.contains,
)


_BOOLEAN_OPERATORS = (
    operator.and_,
    operator.or_,
    operator.xor,
    operator.invert
)


def infer_type_builtins(builtin):
    if builtin in _BINARY_OPERATORS:
        n_params = 2
    elif builtin in _UNARY_OPERATORS:
        n_params = 1
    elif builtin in _NARY_OPERATORS:
        n_params = None
    else:
        try:
            signature = inspect.signature(builtin)
            n_params = len(signature.parameters)
        except ValueError:
            return Callable[..., Unknown]

    if builtin in _BOOLEAN_OPERATORS:
        params_type = [bool] * n_params
        return_type = bool
    elif builtin in _RELATIVE_OPERATORS:
        params_type = [Unknown] * n_params
        return_type = bool
    elif builtin in _NARY_OPERATORS:
        params_type = [Unknown]
        return_type = Unknown
    else:
        params_type = [Unknown] * n_params
        return_type = Unknown

    return Callable[params_type, return_type]


def get_args(type_):
    if is_parameterized(type_):
        ret = type_.__args__
        if ret is None:
            ret = tuple()
    elif is_parametrical(type_):
        ret = tuple()
    else:
        raise ValueError(f"Not {type_} is not a generic type")
    return ret


def get_generic_type(typ):
    return getattr(typ, "__generic_class__", typ)



from ...expression_walker import IdentityWalker
from ...expressions import Symbol
from .. import Conjunction, Disjunction, Negation
from ..expression_processing import TranslateToLogic


class Translator(TranslateToLogic, IdentityWalker):
    pass


def test_translation_negation():
    Q = Symbol('Q')
    x = Symbol('x')

    ttl = Translator()

    exp = ~Q(x)
    res = ttl.walk(exp)

    assert isinstance(res, Negation)
    assert res.formula == Q(x)


def test_translation_conjunctions():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    x = Symbol('x')

    Qx = Q(x)
    Rx = R(x)
    Sx = S(x)

    ttl = Translator()

    conj1 = Q(x) & R(x)
    res = ttl.walk(conj1)
    assert isinstance(res, Conjunction)
    assert res.formulas == (Qx, Rx)

    conj2 = conj1 & S(x)
    res = ttl.walk(conj2)
    assert isinstance(res, Conjunction)
    assert res.formulas == (Qx, Rx, Sx)

    conj3 = S(x) & conj1
    res = ttl.walk(conj3)
    assert isinstance(res, Conjunction)
    assert res.formulas == (Sx, Qx, Rx)


def test_translation_disjunctions():
    Q = Symbol('Q')
    R = Symbol('R')
    S = Symbol('S')
    x = Symbol('x')

    Qx = Q(x)
    Rx = R(x)
    Sx = S(x)

    ttl = Translator()

    disj1 = Q(x) | R(x)
    res = ttl.walk(disj1)
    assert isinstance(res, Disjunction)
    assert res.formulas == (Qx, Rx)

    disj2 = disj1 | S(x)
    res = ttl.walk(disj2)
    assert isinstance(res, Disjunction)
    assert res.formulas == (Qx, Rx, Sx)

    disj3 = S(x) | disj1
    res = ttl.walk(disj3)
    assert isinstance(res, Disjunction)
    assert res.formulas == (Sx, Qx, Rx)



from operator import and_, invert, or_
from typing import Any

from ..exceptions import NeuroLangException
from ..expression_walker import PatternWalker, add_match
from ..expressions import Constant, FunctionApplication, Symbol
from ..utils import OrderedSet
from . import (
    FALSE,
    TRUE,
    Conjunction,
    Disjunction,
    Implication,
    LogicOperator,
    Negation,
    Quantifier,
    Union,
)


class LogicSolver(PatternWalker):
    @add_match(Conjunction)
    def evaluate_conjunction(self, expression):
        unsolved_formulas = tuple()
        for formula in expression.formulas:
            solved_formula = self.walk(formula)
            if isinstance(solved_formula, Constant):
                if not bool(solved_formula.value):
                    return FALSE
            else:
                unsolved_formulas += (solved_formula,)

        if len(unsolved_formulas) == 0:
            return TRUE
        elif len(unsolved_formulas) == 1:
            return unsolved_formulas[0]
        else:
            return Conjunction(unsolved_formulas)

    @add_match(Disjunction)
    def evaluate_disjunction(self, expression):
        unsolved_formulas = tuple()
        for formula in expression.formulas:
            solved_formula = self.walk(formula)
            if isinstance(solved_formula, Constant):
                if bool(solved_formula.value):
                    return TRUE
            else:
                unsolved_formulas += (solved_formula,)

        if len(unsolved_formulas) == 0:
            return FALSE
        elif len(unsolved_formulas) == 1:
            return unsolved_formulas[0]
        else:
            return Disjunction(unsolved_formulas)

    @add_match(Negation)
    def evaluate_negation(self, expression):
        solved_formula = self.walk(expression.formula)
        if isinstance(solved_formula, Constant):
            return Constant[bool](not solved_formula.value)
        if isinstance(solved_formula, Negation):
            solved_formula = solved_formula.formula
        else:
            solved_formula = Negation(solved_formula)
        return solved_formula

    @add_match(Implication(..., TRUE))
    def evaluate_implication_true_antecedent(self, expression):
        return self.walk(expression.consequent)

    @add_match(Implication(..., FALSE))
    def evaluate_implication_false_antecedent(self, expression):
        return TRUE

    @add_match(Implication(TRUE, ...))
    def evaluate_implication_true_consequent(self, expression):
        return TRUE

    @add_match(Implication(FALSE, ...))
    def evaluate_implication_false_consequent(self, expression):
        return self.walk(Negation(expression.antecedent))

    @add_match(Implication)
    def evaluate_implication(self, expression):
        solved_antecedent = self.walk(expression.antecedent)
        if solved_antecedent is not expression.antecedent:
            return self.walk(
                Implication(expression.consequent, solved_antecedent)
            )

        solved_consequent = self.walk(expression.consequent)
        if solved_consequent is not expression.consequent:
            return self.walk(Implication(solved_consequent, solved_antecedent))

        return expression


def is_logic_function_application(function_application):
    if not isinstance(function_application.functor, Constant):
        return False
    functor_value = function_application.functor.value
    return functor_value in (and_, or_, invert)


class TranslateToLogic(PatternWalker):
    @add_match(FunctionApplication(Constant[Any](and_), ...))
    def build_conjunction(self, conjunction):
        args = tuple()
        for arg in conjunction.args:
            new_arg = self.walk(arg)
            if isinstance(new_arg, Conjunction):
                args += new_arg.formulas
            else:
                args += (new_arg,)

        return self.walk(Conjunction(args))

    @add_match(FunctionApplication(Constant[Any](or_), ...))
    def build_disjunction(self, disjunction):
        args = tuple()
        for arg in disjunction.args:
            new_arg = self.walk(arg)
            if isinstance(new_arg, Disjunction):
                args += new_arg.formulas
            else:
                args += (new_arg,)

        return self.walk(Disjunction(args))

    @add_match(FunctionApplication(Constant[Any](invert), ...))
    def build_negation(self, inversion):
        arg = self.walk(inversion.args[0])
        return self.walk(Negation(arg))

    @add_match(
        LogicOperator,
        lambda expression: any(
            isinstance(arg, FunctionApplication)
            and is_logic_function_application(arg)
            for arg in expression.unapply()
        ),
    )
    def translate_logic_operator(self, expression):
        args = expression.unapply()
        new_args = tuple()
        changed = False
        for arg in args:
            if isinstance(
                arg, FunctionApplication
            ) and is_logic_function_application(arg):
                new_arg = self.walk(arg)
            else:
                new_arg = arg

            if new_arg is not arg:
                changed = True

            new_args += (new_arg,)

        if changed:
            expression = self.walk(expression.apply(*new_args))
        return expression


class WalkLogicProgramAggregatingSets(PatternWalker):
    @add_match(Conjunction)
    def conjunction(self, expression):
        fvs = OrderedSet()
        for formula in expression.formulas:
            fvs |= self.walk(formula)
        return fvs

    @add_match(Union)
    def union(self, expression):
        return self.conjunction(expression)

    @add_match(Disjunction)
    def disjunction(self, expression):
        return self.conjunction(expression)

    @add_match(Negation)
    def negation(self, expression):
        return self.walk(expression.formula)

    @add_match(Quantifier)
    def quantifier(self, expression):
        return self.walk(expression.body)

    @add_match(LogicOperator)
    def logic_operator(self, expression):
        fvs = OrderedSet()
        for arg in expression.unapply():
            fvs |= self.walk(arg)
        return fvs


class ExtractFreeVariablesWalker(WalkLogicProgramAggregatingSets):
    @add_match(FunctionApplication)
    def extract_variables_fa(self, expression):
        args = expression.args

        variables = OrderedSet()
        for a in args:
            if isinstance(a, Symbol):
                variables.add(a)
            elif isinstance(a, FunctionApplication):
                variables |= self.walk(a)
            elif isinstance(a, Constant):
                pass
            else:
                raise NeuroLangException("Not a Datalog function application")
        return variables

    @add_match(Quantifier)
    def extract_variables_q(self, expression):
        return self.walk(expression.body) - expression.head._symbols

    @add_match(Implication)
    def extract_variables_s(self, expression):
        return self.walk(expression.antecedent) - self.walk(
            expression.consequent
        )

    @add_match(Symbol)
    def extract_variables_symbol(self, expression):
        return OrderedSet((expression,))

    @add_match(Constant)
    def _(self, expression):
        return OrderedSet()


def extract_logic_free_variables(expression):
    """Extract variables from expression assuming it's in logic format.

    Parameters
    ----------
    expression : Expression


    Returns
    -------
        OrderedSet
            set of all free variables in the expression.
    """
    efvw = ExtractFreeVariablesWalker()
    return efvw.walk(expression)


class ExtractLogicPredicates(WalkLogicProgramAggregatingSets):
    @add_match(Symbol)
    def symbol(self, expression):
        return OrderedSet()

    @add_match(Constant)
    def constant(self, expression):
        return OrderedSet()

    @add_match(FunctionApplication)
    def extract_predicates_fa(self, expression):
        return OrderedSet([expression])

    @add_match(Negation)
    def negation(self, expression):
        return OrderedSet([expression])


def extract_logic_predicates(expression):
    """Extract predicates from expression
    knowing that it's in logic format, if
    predicates are negated then (not predicate)
    will be the result.ormat

    Parameters
    ----------
    expression : Expression
        expression to extract predicates from


    Returns
    -------
    OrderedSet
        set of all predicates in the expression in lexicographical
        order.

    """
    edp = ExtractLogicPredicates()
    return edp.walk(expression)


class ExtractLogicAtoms(WalkLogicProgramAggregatingSets):
    @add_match(Symbol)
    def symbol(self, expression):
        return OrderedSet()

    @add_match(Constant)
    def constant(self, expression):
        return OrderedSet()

    @add_match(FunctionApplication)
    def extract_predicates_fa(self, expression):
        return OrderedSet([expression])


def extract_logic_atoms(expression):
    """Extract atoms from expression
    knowing that it's in logic format

    Parameters
    ----------
    expression : Expression
        expression to extract predicates from


    Returns
    -------
    OrderedSet
        set of all atoms in the expression in lexicographical
        order.

    """
    edp = ExtractLogicAtoms()
    return edp.walk(expression)



import pytest

from ...expressions import Constant, Symbol
from ...logic import Conjunction, Implication
from ...relational_algebra import ColumnStr, NamedRelationalAlgebraFrozenSet
from ...relational_algebra_provenance import ProvenanceAlgebraSet
from .. import dichotomy_theorem_based_solver, weighted_model_counting
from ..cplogic import testing
from ..cplogic.program import CPLogicProgram
from ..expressions import PROB, Condition, ProbabilisticQuery

ans = Symbol("ans")
P = Symbol("P")
Q = Symbol("Q")
R = Symbol("R")
Z = Symbol("Z")
H = Symbol("H")
A = Symbol("A")
B = Symbol("B")
C = Symbol("C")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")

a = Constant("a")
b = Constant("b")
c = Constant("c")


@pytest.fixture(
    params=((weighted_model_counting, dichotomy_theorem_based_solver)),
    ids=["SDD-WMC", "dichotomy-Safe query"],
)
def solver(request):
    return request.param


def test_marg_query_ground_conditioning(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [
            (0.2, "a", "a"),
            (0.4, "b", "a"),
            (0.1, "b", "c"),
            (0.9, "c", "a"),
        ],
    )
    cpl.add_probabilistic_choice_from_tuples(
        Z,
        [
            (0.2, "c"),
            (0.5, "b"),
            (0.3, "a"),
        ],
    )
    query = Implication(
        Q(x, ProbabilisticQuery(PROB, (x,))), Condition(Z(x), R(x, a))
    )
    cpl.walk(query)
    result = solver.solve_marg_query(query, cpl)
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x"),
            [
                (0.5, "b"),
                (0.3, "a"),
                (0.2, "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_marg_query_two_vars_conditioning(solver):
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [
            (0.2, "a", "a"),
            (0.4, "b", "a"),
            (0.1, "b", "c"),
            (0.9, "c", "a"),
            (0.1, "c", "b"),
        ],
    )
    cpl.add_probabilistic_choice_from_tuples(
        Z,
        [
            (0.2, "c"),
            (0.5, "b"),
            (0.3, "a"),
        ],
    )
    query = Implication(
        Q(x, y, ProbabilisticQuery(PROB, (x, y))), Condition(Z(x), R(x, y))
    )
    cpl.walk(query)
    result = solver.solve_marg_query(query, cpl)
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.3, "a", "a"),
                (0.5, "b", "a"),
                (0.5, "b", "c"),
                (0.2, "c", "a"),
                (0.2, "c", "b"),
            ],
        ),
        ColumnStr("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)


def test_marg_query_conjunctive_conditioned_and_conditioning(solver):
    if solver != dichotomy_theorem_based_solver:
        return
    cpl = CPLogicProgram()
    cpl.add_probabilistic_facts_from_tuples(
        R,
        [
            (0.2, "a", "a"),
            (0.4, "b", "a"),
            (0.1, "b", "c"),
            (0.9, "c", "a"),
            (0.1, "c", "b"),
        ],
    )
    cpl.add_probabilistic_choice_from_tuples(
        Z,
        [
            (0.2, "c"),
            (0.5, "b"),
            (0.3, "a"),
        ],
    )
    cpl.add_probabilistic_choice_from_tuples(
        A,
        [
            (0.9, "b"),
            (0.1, "a"),
        ],
    )
    cpl.add_probabilistic_facts_from_tuples(
        B,
        [
            (0.3, "b"),
            (0.9, "a"),
        ],
    )
    cpl.add_probabilistic_facts_from_tuples(
        C,
        [
            (0.01, "a"),
            (0.27, "c"),
        ],
    )
    cpl.walk(Implication(H(x, y), Conjunction((Z(x), R(y, x)))))
    # Prob{ H(x, y) ^ C(x) | A(x) ^ B(x) }
    query = Implication(
        Q(x, y, ProbabilisticQuery(PROB, (x, y))),
        Condition(Conjunction((H(x, y), C(x))), Conjunction((A(x), B(x)))),
    )
    cpl.walk(query)
    result = solver.solve_marg_query(query, cpl)
    expected = ProvenanceAlgebraSet(
        NamedRelationalAlgebraFrozenSet(
            ("_p_", "x", "y"),
            [
                (0.2 * 0.01 * 0.3, "a", "a"),
                (0.4 * 0.01 * 0.3, "a", "b"),
                (0.9 * 0.01 * 0.3, "a", "c"),
            ],
        ),
        ColumnStr("_p_"),
    )
    assert testing.eq_prov_relations(result, expected)



from .pandas import (NamedRelationalAlgebraFrozenSet,
                     RelationalAlgebraFrozenSet, RelationalAlgebraSet,
                     RelationalAlgebraColumnInt, RelationalAlgebraColumnStr,
                     RelationalAlgebraStringExpression)

__all__ = [
    "RelationalAlgebraColumnInt",
    "RelationalAlgebraColumnStr",
    "RelationalAlgebraStringExpression",
    "RelationalAlgebraFrozenSet",
    "RelationalAlgebraSet",
    "NamedRelationalAlgebraFrozenSet"
]



r"""
Probabilistic Frontend
======================
Complements QueryBuilderDatalog class with probabilistic capabilities
1- add extensional probabilistic facts and choices
2- sove probabilistic queries
"""
import collections
import typing
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Type
)
from uuid import uuid1

from .. import expressions as ir
from ..datalog.aggregation import (
    Chase,
    DatalogWithAggregationMixin,
    TranslateToLogicWithAggregation
)
from ..datalog.constraints_representation import DatalogConstraintsProgram
from ..datalog.negation import DatalogProgramNegationMixin
from ..datalog.ontologies_parser import OntologyParser
from ..datalog.ontologies_rewriter import OntologyRewriter
from ..exceptions import UnsupportedQueryError
from ..expression_walker import ExpressionBasicEvaluator
from ..logic import Union
from ..probabilistic.cplogic.program import (
    CPLogicMixin,
    TranslateProbabilisticQueryMixin
)
from ..probabilistic.dichotomy_theorem_based_solver import (
    solve_marg_query as lifted_solve_marg_query,
    solve_succ_query as lifted_solve_succ_query
)
from ..probabilistic.expression_processing import (
    is_probabilistic_predicate_symbol,
    is_within_language_prob_query
)
from ..probabilistic.query_resolution import compute_probabilistic_solution
from ..probabilistic.stratification import stratify_program
from ..region_solver import RegionSolver
from ..relational_algebra import (
    NamedRelationalAlgebraFrozenSet,
    RelationalAlgebraStringExpression
)
from . import query_resolution_expressions as fe
from .datalog.syntax_preprocessing import ProbFol2DatalogMixin
from .query_resolution_datalog import QueryBuilderDatalog


class RegionFrontendCPLogicSolver(
    TranslateProbabilisticQueryMixin,
    TranslateToLogicWithAggregation,
    ProbFol2DatalogMixin,
    RegionSolver,
    CPLogicMixin,
    DatalogWithAggregationMixin,
    DatalogProgramNegationMixin,
    DatalogConstraintsProgram,
    ExpressionBasicEvaluator,
):
    pass


class NeurolangPDL(QueryBuilderDatalog):
    """
    Complements QueryBuilderDatalog class with probabilistic capabilities
    1- add extensional probabilistic facts and choices
    2- sove probabilistic queries
    """

    def __init__(
        self,
        chase_class: Type[Chase] = Chase,
        probabilistic_solver: Callable = lifted_solve_succ_query,
        probabilistic_marg_solver: Callable = lifted_solve_marg_query,
    ) -> "ProbabilisticFrontend":
        """
        Query builder with probabilistic capabilities

        Parameters
        ----------
        chase_class : Type[Chase], optional
            used to compute deterministic solutions, by default Chase
        probabilistic_solver : Callable, optional
            used to compute probabilistic solutions,
            by default lifted_solve_succ_query

        Returns
        -------
        ProbabilisticFrontend
            see description
        """
        super().__init__(
            RegionFrontendCPLogicSolver(), chase_class=chase_class
        )
        self.probabilistic_solver = probabilistic_solver
        self.probabilistic_marg_solver = lifted_solve_marg_query
        self.ontology_loaded = False

    def load_ontology(
        self,
        paths: typing.Union[str, List[str]],
        load_format: typing.Union[str, List[str]] = "xml",
    ) -> None:
        """
        Loads and parses ontology stored at the specified paths, and
        store them into attributes

        Parameters
        ----------
        paths : typing.Union[str, List[str]]
            where the ontology files are stored
        load_format : typing.Union[str, List[str]], optional
            storage format, by default "xml"
        """
        onto = OntologyParser(paths, load_format)
        d_pred, u_constraints = onto.parse_ontology()
        self.program_ir.walk(u_constraints)
        self.program_ir.add_extensional_predicate_from_tuples(
            onto.get_triples_symbol(), d_pred[onto.get_triples_symbol()]
        )
        self.program_ir.add_extensional_predicate_from_tuples(
            onto.get_pointers_symbol(), d_pred[onto.get_pointers_symbol()]
        )

    @property
    def current_program(self) -> List[fe.Expression]:
        """
        Returns the list of Front End Expressions that have
        currently been declared in the program, or through
        the program's constraints

        Returns
        -------
        List[fe.Expression]
            see description

        Example
        -------
        >>> nl = ProbabilisticFrontend()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        ...     cp = nl.current_program
        >>> cp
        [
            Z(x, ( PROB(x) )) ← ( P(x) ) ∧ ( Q(x) )
        ]
        """
        cp = []
        for constraint in self.program_ir.constraints().formulas:
            cp.append(self.frontend_translator.walk(constraint))
        cp += super().current_program
        return cp

    def _execute_query(
        self,
        head: typing.Union[fe.Symbol, Tuple[fe.Expression, ...]],
        predicate: fe.Expression,
    ) -> Tuple[AbstractSet, Optional[ir.Symbol]]:
        """
        [Internal usage - documentation for developpers]

        Performs an inferential query: will return as first output
        an AbstractSet with as many elements as solutions
        of the predicate query. AbstractSet's columns correspond to
        the expressions in the head.
        Typically, probabilities are abstracted and processed similar
        to symbols, though of different nature (see examples)
        If head expressions are arguments of a functor, the latter will
        be returned as the second output, defaulted as None

        Parameters
        ----------
        head : typing.Union[
            fe.Symbol,
            Tuple[fe.Expression, ...],
        ]
            see description
        predicate : Expression
            see description

        Returns
        -------
        Tuple[AbstractSet, Optional[ir.Symbol]]
            see description

        Examples
        --------
        Note: example ran with pandas backend
        >>> nl = ProbabilisticFrontend()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        ...     s1 = nl._execute_query(tuple(), e.Z[e.x, e.p])
        ...     s2 = nl._execute_query((e.x, ), e.Z[e.x, e.p])
        ...     s3 = nl._execute_query(e.Z[e.x, e.p], e.Z[e.x, e.p])
        >>> s1
        (
            C{
                Empty DataFrame
                Columns: []
                Index: [0, 1]
                : typing.AbstractSet
            },
            None
        )
        >>> s2
        (
            C{
                    x
                0   a
                1   c
                : typing.AbstractSet
            },
            None
        )
        >>> s3
        (
            C{
                    x   p
                0   a   0.111111
                1   c   0.111111
                : typing.AbstractSet
            },
            S{Z: Unknown}
        )
        """
        query_pred_symb = predicate.expression.functor
        if is_probabilistic_predicate_symbol(query_pred_symb, self.program_ir):
            raise UnsupportedQueryError(
                "Queries on probabilistic predicates are not supported"
            )
        query = self.program_ir.symbol_table[query_pred_symb].formulas[0]
        solution = self._solve(query)
        if not isinstance(head, tuple):
            # assumes head is a predicate e.g. r(x, y)
            head_symbols = tuple(head.expression.args)
            functor_orig = head.expression.functor
        else:
            head_symbols = tuple(t.expression for t in head)
            functor_orig = None
        solution = self._restrict_to_query_solution(
            head_symbols, predicate, solution
        )
        return solution, functor_orig

    def solve_all(self) -> Dict[str, NamedRelationalAlgebraFrozenSet]:
        """
        Returns a dictionary of "predicate_name": "Content"
        for all elements in the solution of the Datalog program.
        Typically, probabilities are abstracted and processed similar
        to symbols, though of different nature (see examples)

        Returns
        -------
        Dict[str, NamedRelationalAlgebraFrozenSet]
            extensional and intentional facts that have been derived
            through the current program, optionally with probabilities

        Example
        -------
        Note: example ran with pandas backend
        >>> nl = ProbabilisticFrontend()
        >>> P = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("b",), ("c",)], name="P"
        ... )
        >>> Q = nl.add_uniform_probabilistic_choice_over_set(
        ...     [("a",), ("d",), ("c",)], name="Q"
        ... )
        >>> with nl.scope as e:
        ...     e.Z[e.PROB[e.x], e.x] = P[e.x] & Q[e.x]
        ...     solution = nl.solve_all()
        >>> solution
        {
            'Z':
                PROB        x
            0   0.111111    a
            1   0.111111    c
        }
        """
        solution_ir = self._solve()
        solution = {}
        for k, v in solution_ir.items():
            solution[k.name] = NamedRelationalAlgebraFrozenSet(
                self.predicate_parameter_names(k.name), v.value.unwrap()
            )
            solution[k.name].row_type = v.value.row_type
        return solution

    def _solve(self, query=None):
        idbs = stratify_program(query, self.program_ir)
        det_idb = idbs.get("deterministic", Union(tuple()))
        prob_idb = idbs.get("probabilistic", Union(tuple()))
        postprob_idb = idbs.get("post_probabilistic", Union(tuple()))
        solution = self._solve_deterministic_stratum(det_idb)
        if prob_idb.formulas:
            solution = self._solve_probabilistic_stratum(solution, prob_idb)
        if postprob_idb.formulas:
            solution = self._solve_postprobabilistic_deterministic_stratum(
                solution, postprob_idb
            )
        return solution

    def _solve_deterministic_stratum(self, det_idb):
        if "__constraints__" in self.symbol_table:
            eB = self._rewrite_program_with_ontology(det_idb)
            det_idb = Union(det_idb.formulas + eB.formulas)
        chase = self.chase_class(self.program_ir, rules=det_idb)
        solution = chase.build_chase_solution()
        return solution

    def _solve_probabilistic_stratum(self, solution, prob_idb):
        pfact_edb = self.program_ir.probabilistic_facts()
        pchoice_edb = self.program_ir.probabilistic_choices()
        prob_solution = compute_probabilistic_solution(
            solution,
            pfact_edb,
            pchoice_edb,
            prob_idb,
            self.probabilistic_solver,
            self.probabilistic_marg_solver,
        )
        wlq_symbs = set(
            rule.consequent.functor
            for rule in prob_idb.formulas
            if is_within_language_prob_query(rule)
        )
        solution.update(
            {
                pred_symb: relation
                for pred_symb, relation in prob_solution.items()
                if pred_symb in wlq_symbs
            }
        )
        return solution

    def _solve_postprobabilistic_deterministic_stratum(
        self, solution, postprob_idb
    ):
        solver = RegionFrontendCPLogicSolver()
        for psymb, relation in solution.items():
            solver.add_extensional_predicate_from_tuples(
                psymb,
                relation.value,
            )
        for builtin_symb in self.program_ir.builtins():
            solver.symbol_table[builtin_symb] = self.program_ir.symbol_table[
                builtin_symb
            ]
        solver.walk(postprob_idb)
        chase = self.chase_class(solver, rules=postprob_idb)
        solution = chase.build_chase_solution()
        return solution

    @staticmethod
    def _restrict_to_query_solution(head_symbols, predicate, solution):
        """
        Based on a solution instance and a query predicate, retrieve the
        relation whose columns correspond to symbols in the head of the query.
        """
        pred_symb = predicate.expression.functor
        # return dum when empty solution (reported in GH481)
        if pred_symb not in solution:
            return ir.Constant[AbstractSet](
                NamedRelationalAlgebraFrozenSet.dum()
            )
        query_solution = solution[pred_symb].value.unwrap()
        cols = list(
            arg.name
            for arg in predicate.expression.args
            if isinstance(arg, ir.Symbol)
        )
        query_solution = NamedRelationalAlgebraFrozenSet(cols, query_solution)
        query_solution = query_solution.projection(
            *(symb.name for symb in head_symbols)
        )
        return ir.Constant[AbstractSet](query_solution.to_unnamed())

    def _rewrite_program_with_ontology(self, deterministic_program):
        orw = OntologyRewriter(
            deterministic_program, self.program_ir.constraints()
        )
        rewrite = orw.Xrewrite()

        eB = ()
        for imp in rewrite:
            eB += (imp[0],)

        return Union(eB)

    def add_probabilistic_facts_from_tuples(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add probabilistic facts from tuples whose first element
        contains the probability label attached to that tuple.
        In the tuple (p, a, b, ...), p is the float probability
        of tuple (a, b, ...) to be True in any possible world.

        Note that each tuple from the iterable is independant
        from the others, meaning that multiple tuples can be True
        in the same possible world, contrary to a probabilistic choice.
        See example for details.

        Warning
        -------
        Typing for the iterable is improper, true -but yet unsupported
        in Python typing- typing should be Iterable[Tuple[float, Any, ...]]
        See examples

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            the first float number represents the probability
            of the tuple constituted of the remaining elements
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Example
        -------
        >>> nl = ProbabilisticFrontend()
        >>> p = [(0.8, 'a', 'b'), (0.7, 'b', 'c')]
        >>> nl.add_probabilistic_facts_from_tuples(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str]] = \
            [(0.8, 'a', 'b'), (0.7, 'b', 'c')]

        Adds the probabilistic facts:
            P((a, b)) : 0.8  <-  T
            P((b, c)) : 0.7  <-  T
        """
        return self._add_probabilistic_tuples(
            iterable,
            type_,
            name,
            self.program_ir.add_probabilistic_facts_from_tuples,
        )

    def add_probabilistic_choice_from_tuples(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add probabilistic choice from tuples whose first element
        contains the probability label attached to that tuple.
        In the tuple (p, a, b, ...), p is the float probability
        of tuple (a, b, ...) to be True and all remaining tuples to
        be False in any possible world.

        Note that, contrary to a list of probabilistic facts, this represents
        a choice among possible values for the predicate, meaning that tuples
        in the set are mutually exclusive.
        See example for details.

        Warning
        -------
        Typing for the iterable is improper, true -but yet unsupported-
        typing should be Iterable[Tuple[float, Any, ...]]
        See examples

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            the first float number represents the probability
            of the tuple constituted of the remaining elements
            Note that the float probabilities must sum to 1.
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Raises
        ------
        DistributionDoesNotSumToOneError
            if float probabilities do not sum to 1.

        Example
        -------
        >>> nl = ProbabilisticFrontend()
        >>> p = [(0.8, 'a', 'b'), (0.2, 'b', 'c')]
        >>> nl.add_probabilistic_choice_from_tuples(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str, str]] = \
            [(0.8, 'a', 'b'), (0.2, 'b', 'c')]

        Adds the probabilistic choice:
            P((a, b)) : 0.8  v  P((b, c)) : 0.2  <-  T
        """
        return self._add_probabilistic_tuples(
            iterable,
            type_,
            name,
            self.program_ir.add_probabilistic_choice_from_tuples,
        )

    def _add_probabilistic_tuples(
        self, iterable, type_, name, solver_add_method
    ):
        if name is None:
            name = str(uuid1())
        if isinstance(type_, tuple):
            type_ = Tuple[type_]
        symbol = ir.Symbol[AbstractSet[type_]](name)
        solver_add_method(symbol, iterable)
        return fe.Symbol(self, name)

    def add_uniform_probabilistic_choice_over_set(
        self,
        iterable: Iterable[Tuple[Any, ...]],
        type_: Type = ir.Unknown,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Add uniform probabilistic choice over values in the iterable.

        Every tuple in the iterable will be assigned the same probability to
        be True, with all remaning tuples False, in any possible world.

        Note that, contrary to a list of probabilistic facts, this represents
        a choice among possible values for the predicate, meaning that tuples
        in the set are mutually exclusive.

        See example for details.

        Parameters
        ----------
        iterable : Iterable[Tuple[Any, ...]]
            typically List[Tuple[Any, ...]], other types of Iterable
            will be cast as lists
        type_ : Type, optional
            type for resulting AbstractSet if None will be inferred
            from the data, by default ir.Unknown
        name : Optional[str], optional
            name for the resulting fe.Symbol, if None
            will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description

        Example
        -------
        >>> nl = ProbabilisticFrontend()
        >>> p = [('a',), ('b',)]
        >>> nl.add_uniform_probabilistic_choice_over_set(p, name="P")
        P: typing.AbstractSet[typing.Tuple[float, str]] = \
            [(0.5, 'a'), (0.5, 'b')]

        Adds the probabilistic choice:
            P(a) : 0.5  v  P(b) : 0.5  <-  T
        """
        if name is None:
            name = str(uuid1())
        if isinstance(type_, tuple):
            type_ = Tuple[type_]
        symbol = ir.Symbol[AbstractSet[type_]](name)
        arity = len(next(iter(iterable)))
        columns = tuple(ir.Symbol.fresh().name for _ in range(arity))
        ra_set = NamedRelationalAlgebraFrozenSet(columns, iterable)
        prob_col = ir.Symbol.fresh().name
        probability = 1 / len(iterable)
        projections = collections.OrderedDict()
        projections[prob_col] = probability
        for col in columns:
            projections[col] = RelationalAlgebraStringExpression(col)
        ra_set = ra_set.extended_projection(projections)
        self.program_ir.add_probabilistic_choice_from_tuples(symbol, ra_set)
        return fe.Symbol(self, name)


ProbabilisticFrontend = NeurolangPDL



from collections.abc import Iterable, Mapping, MutableMapping, MutableSet, Set
from typing import AbstractSet, Tuple

from ..expression_walker import ReplaceExpressionsByValues
from ..expressions import Constant, FunctionApplication
from ..type_system import Unknown, get_args
from ..utils.relational_algebra_set import RelationalAlgebraFrozenSet
from .wrapped_collections import (WrappedRelationalAlgebraFrozenSet,
                                  WrappedRelationalAlgebraSet)


class FrozenInstance:
    _set_type = WrappedRelationalAlgebraFrozenSet
    _rebv = ReplaceExpressionsByValues({})

    def __init__(self, elements=None):
        self.cached_hash = None
        if elements is None:
            elements = dict()
        elif isinstance(elements, FrozenInstance):
            self.cached_hash = elements.cached_hash
            elements = elements.elements.copy()
        elif isinstance(elements, Mapping):
            elements = self._elements_from_mapping(elements)
        elif isinstance(elements, Iterable):
            elements = self._elements_from_iterable(elements)
        self.elements = elements

    def _elements_from_mapping(self, elements):
        in_elements = elements
        elements = dict()
        for k, v in in_elements.items():
            v = self._get_set(v)
            if self._set_not_empty(v):
                elements[k] = v
        return elements

    def _get_set(self, v):
        v_type = Unknown
        if not isinstance(v, self._set_type):
            v, v_type = self._get_set_and_type(v)
            if self._set_not_empty(v):
                v = self._set_type(
                    v, row_type=v_type, verify_row_type=False
                )
        return v

    def _set_not_empty(self, v):
        return (
            (isinstance(v, RelationalAlgebraFrozenSet) and not v.is_empty())
            or len(v) > 0
        )

    def _get_set_and_type(self, v):
        row_type = Unknown
        if isinstance(v, Constant[AbstractSet[Tuple]]):
            row_type = get_args(v.type)[0]
            if isinstance(v.value, self._set_type):
                v = v.value
            else:
                v = self._rebv.walk(v)
        return v, row_type

    def _is_expression_iterable(self, v):
        is_expression = False
        for element in v:
            if isinstance(element, Constant):
                is_expression = True
            elif (
                isinstance(element, tuple) and
                isinstance(element[0], Constant)
            ):
                is_expression = True
            break
        return is_expression

    def _elements_from_iterable(self, iterable):
        result = dict()
        for predicate in iterable:
            symbol = predicate.functor
            if symbol not in result:
                result[symbol] = []
            result[symbol].append(self._rebv.walk(predicate.args))

        for symbol in result:
            result[symbol] = self._set_type(result[symbol])
        return result

    def __hash__(self):
        if self.cached_hash is None:
            self.cached_hash = hash(tuple(zip(self.elements.items())))
        return self.cached_hash

    def __or__(self, other):
        if not isinstance(other, Instance):
            return super().__or__(other)
        new_elements = dict()
        for predicate in (self.elements.keys() | other.elements.keys()):
            new_elements[predicate] = self._set_type(
                self.elements.get(predicate, self._set_type()) |
                other.elements.get(predicate, self._set_type())
            )
        res = type(self)(new_elements)
        return res

    def __sub__(self, other):
        if not isinstance(other, Instance):
            return super().__sub__(other)
        new_elements = dict()
        for predicate, tuple_set in self.elements.items():
            new_set = tuple_set - other.elements.get(predicate, set())
            if not new_set.is_empty():
                new_elements[predicate] = new_set
        res = type(self)(new_elements)
        return res

    def __and__(self, other):
        if not isinstance(other, Instance):
            return super().__and__(other)

        new_elements = dict()
        for predicate in (self.elements.keys() & other.elements.keys()):
            new_set = self.elements[predicate] & other.elements[predicate]
            if not new_set.is_empty():
                new_elements[predicate] = new_set
        res = type(self)(new_elements)
        return res

    def copy(self):
        new_copy = type(self)()
        new_copy.elements = self.elements.copy()
        new_copy.hash = self.cached_hash
        return new_copy

    def _create_view(self, class_):
        out = class_()
        out.elements = self.elements
        return out

    def __eq__(self, other):
        if isinstance(other, Instance):
            return self.elements == other.elements
        else:
            return super().__eq__(other)

    def __repr__(self):
        return repr(self.elements)


class FrozenMapInstance(FrozenInstance, Mapping):
    def _set_to_constant(self, set_, type_=Unknown):
        if type_ is Unknown:
            type_ = set_.row_type
        return Constant[AbstractSet[type_]](
            set_,
            verify_type=False
        )

    def __getitem__(self, predicate_symbol):
        set_ = self.elements[predicate_symbol]
        type_ = set_.row_type
        return self._set_to_constant(set_, type_=type_)

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)

    def items(self):
        for k, v in self.elements.items():
            row_type = v.row_type
            yield k, self._set_to_constant(
                v, type_=row_type
            )

    def values(self):
        for k, v in self.elements.items():
            row_type = v.row_type
            yield self._set_to_constant(v, type_=row_type)

    def as_set(self):
        return self._create_view(FrozenSetInstance)

    def as_map(self):
        return self

    def __hash__(self):
        return super().__hash__()


class FrozenSetInstance(FrozenInstance, Set):
    def __contains__(self, predicate):
        functor = predicate.functor
        tuple_values = predicate.args
        return (
            functor in self.elements and
            tuple_values in self.elements[functor]
        )

    def __len__(self):
        return sum(len(v) for v in self.elements.values())

    def __iter__(self):
        for predicate, tuples in self.elements.items():
            for t in tuples:
                arg = t.value
                yield FunctionApplication(
                    predicate, arg
                )

    def as_set(self):
        return self

    def as_map(self):
        return self._create_view(FrozenMapInstance)

    def __hash__(self):
        return super().__hash__()


class Instance(FrozenInstance):
    _set_type = WrappedRelationalAlgebraSet

    def __init__(self, elements=None):
        super().__init__(elements=elements)

    def __hash__(self):
        raise TypeError('Instance objects are mutable and cannot be hashed')

    def __ior__(self, other):
        if not isinstance(other, Instance):
            return super().__ior__(other)

        for predicate in (self.elements.keys() & other.elements.keys()):
            self.elements[predicate] |= other.elements[predicate]
        for predicate in (other.elements.keys() - self.elements.keys()):
            self.elements[predicate] = other.elements[predicate]
        return self

    def _remove_predicate_symbol(self, predicate_symbol):
        del self.elements[predicate_symbol]

    def __isub__(self, other):
        if not isinstance(other, Instance):
            return super().__isub__(other)
        for predicate in (self.elements.keys() & other.elements.keys()):
            self.elements[predicate] -= other.elements[predicate]
            if self.elements[predicate].is_empty():
                self._remove_predicate_symbol(predicate)
        return self

    def __iand__(self, other):
        if isinstance(other, Instance):
            return super().__iand__(other)
        subs = self.elements.keys() - other.elements.keys()
        for predicate in subs:
            self._remove_predicate_symbol(predicate)
        for predicate in self.elements:
            self.elements[predicate] &= other.elements[predicate]
            if self.elements[predicate].is_empty():
                self._remove_predicate_symbol(predicate)
        return self

    def copy(self):
        new_copy = type(self)()
        new_copy.elements = self.elements.copy()
        return new_copy


class MapInstance(Instance, FrozenMapInstance, MutableMapping):
    def __setitem__(self, predicate_symbol, value):
        set_ = self._set_type(value.value)
        self.elements[predicate_symbol] = set_

    def __delitem__(self, predicate_symbol):
        self._remove_predicate_symbol(predicate_symbol)

    def as_set(self):
        return self._create_view(SetInstance)

    def as_map(self):
        return self


class SetInstance(Instance, FrozenSetInstance, MutableSet):
    def add(self, value):
        functor = value.functor
        args = value.args
        if functor not in self.elements:
            self.elements[functor] = self._set_type()
        self.elements[value.functor].add(args)

    def discard(self, value):
        functor = value.functor
        value = value.args
        self.elements[functor].discard(value)
        if self.elements[functor].is_empty():
            self._remove_predicate_symbol(functor)

    def as_set(self):
        return self

    def as_map(self):
        return self._create_view(MapInstance)



import tatsu

from .ast import ASTNode


class TatsuASTConverter(object):
    @staticmethod
    def _default(ast):
        if isinstance(ast, tatsu.ast.AST):
            return ASTNode(
                ast['parseinfo'].rule,
                ast
            )
        else:
            return ast
